from __future__ import annotations

from pathlib import Path

from orchesis.policy_engine import PolicyEngine
from orchesis.tool_policy import ToolPolicyEngine


def _base_policy() -> dict:
    return {
        "version": "1.0",
        "name": "production-policy",
        "default_action": "deny",
        "variables": {
            "trusted_agents": ["openclaw", "cursor"],
            "sensitive_tools": ["bash", "exec", "eval"],
        },
        "rules": [
            {
                "id": "block-sensitive-tools",
                "description": "Block dangerous shell tools",
                "when": {"tool": {"in": "$sensitive_tools"}},
                "action": "block",
                "priority": 100,
            },
            {
                "id": "allow-trusted-agents",
                "description": "Trusted agents get broader access",
                "when": {
                    "agent_id": {"in": "$trusted_agents"},
                    "tool": {"not_in": "$sensitive_tools"},
                },
                "action": "allow",
                "priority": 90,
            },
            {
                "id": "rate-limit-search",
                "description": "Limit search calls per session",
                "when": {"tool": {"eq": "web_search"}, "session_calls": {"gt": 10}},
                "action": "block",
                "priority": 80,
            },
            {
                "id": "warn-on-file-write",
                "description": "Warn when writing files",
                "when": {"tool": {"matches": "write_*"}},
                "action": "warn",
                "priority": 70,
            },
        ],
    }


def test_from_dict_basic() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    assert len(engine.get_rules()) == 4


def test_from_yaml(tmp_path: Path) -> None:
    path = tmp_path / "policy.yaml"
    path.write_text(
        """
version: "1.0"
name: "production-policy"
default_action: deny
rules:
  - id: "r1"
    description: "desc"
    when:
      tool: {eq: "bash"}
    action: block
    priority: 10
""".strip()
        + "\n",
        encoding="utf-8",
    )
    engine = PolicyEngine.from_yaml(str(path))
    assert len(engine.get_rules()) == 1


def test_evaluate_eq() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    result = engine.evaluate(tool_name="web_search", session_calls=11)
    assert result.action == "block"


def test_evaluate_in_list() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    result = engine.evaluate(tool_name="bash")
    assert result.action == "block"


def test_evaluate_in_variable() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    result = engine.evaluate(tool_name="eval")
    assert result.action == "block"


def test_evaluate_not_in() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    result = engine.evaluate(tool_name="read_file", agent_id="openclaw")
    assert result.action == "allow"


def test_evaluate_matches_glob() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    result = engine.evaluate(tool_name="write_file")
    assert result.action == "warn"


def test_evaluate_startswith() -> None:
    policy = {
        "default_action": "deny",
        "rules": [
            {"id": "r", "description": "d", "when": {"tool": {"startswith": "file"}}, "action": "allow", "priority": 1}
        ],
    }
    engine = PolicyEngine.from_dict(policy)
    result = engine.evaluate(tool_name="file_read")
    assert result.action == "allow"


def test_evaluate_gt_session_calls() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    result = engine.evaluate(tool_name="web_search", session_calls=12)
    assert result.action == "block"


def test_evaluate_lte() -> None:
    policy = {
        "default_action": "deny",
        "rules": [
            {"id": "r", "description": "d", "when": {"token_count": {"lte": 1000}}, "action": "allow", "priority": 1}
        ],
    }
    engine = PolicyEngine.from_dict(policy)
    result = engine.evaluate(tool_name="read_file", token_count=500)
    assert result.action == "allow"


def test_evaluate_multi_condition_and() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    ok = engine.evaluate(tool_name="read_file", agent_id="openclaw")
    no = engine.evaluate(tool_name="bash", agent_id="openclaw")
    assert ok.action == "allow"
    assert no.action == "block"


def test_evaluate_priority_order() -> None:
    policy = {
        "default_action": "deny",
        "rules": [
            {"id": "low", "description": "d", "when": {"tool": {"contains": "write"}}, "action": "warn", "priority": 1},
            {"id": "high", "description": "d", "when": {"tool": {"contains": "write"}}, "action": "block", "priority": 10},
        ],
    }
    engine = PolicyEngine.from_dict(policy)
    result = engine.evaluate(tool_name="write_file")
    assert result.action == "block"
    assert result.rule_id == "high"


def test_evaluate_default_action() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    result = engine.evaluate(tool_name="unknown")
    assert result.action == "block"
    assert result.matched is False


def test_evaluate_no_match_allow_default() -> None:
    policy = {"default_action": "allow", "rules": []}
    engine = PolicyEngine.from_dict(policy)
    result = engine.evaluate(tool_name="anything")
    assert result.action == "allow"
    assert result.matched is False


def test_variables_resolved() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    vars_map = engine.get_variables()
    assert "sensitive_tools" in vars_map
    assert "bash" in vars_map["sensitive_tools"]


def test_variables_not_found() -> None:
    policy = {
        "default_action": "deny",
        "rules": [
            {"id": "x", "description": "x", "when": {"tool": {"in": "$missing_var"}}, "action": "allow", "priority": 1}
        ],
    }
    engine = PolicyEngine.from_dict(policy)
    result = engine.evaluate(tool_name="bash")
    assert result.matched is False
    assert result.action == "block"


def test_integration_with_tool_policy_engine() -> None:
    pe = PolicyEngine.from_dict(_base_policy())
    tpe = ToolPolicyEngine({"default_action": "allow", "rules": {}}, policy_engine=pe)
    decision = tpe.evaluate(tool_name="bash", agent_id="x", session_id="s1")
    assert decision.action == "block"
    assert decision.rule_source == "policy_engine"


def test_tool_policy_fallthrough() -> None:
    pe = PolicyEngine.from_dict({"default_action": "deny", "rules": []})
    tpe = ToolPolicyEngine({"default_action": "allow", "rules": {"read_file": "allow"}}, policy_engine=pe)
    decision = tpe.evaluate(tool_name="read_file", agent_id="x", session_id="s1")
    assert decision.action == "allow"
    assert decision.rule_source == "explicit_rule"


def test_policy_rule_description_in_result() -> None:
    engine = PolicyEngine.from_dict(_base_policy())
    result = engine.evaluate(tool_name="bash")
    assert result.rule_description == "Block dangerous shell tools"


def test_empty_policy() -> None:
    engine = PolicyEngine.from_dict({})
    result = engine.evaluate(tool_name="read")
    assert result.action == "block"
    assert result.matched is False




from __future__ import annotations

from pathlib import Path

import pytest

from orchesis.config import PolicyError, load_policy
from orchesis.engine import evaluate


def _decision(policy: dict, tool: str, params: dict | None = None):
    return evaluate({"tool": tool, "params": params or {}}, policy)


def test_default_action_deny_blocks_unlisted_tool() -> None:
    policy = {"default_action": "deny", "capabilities": [], "rules": []}
    decision = _decision(policy, "read_file", {"path": "/workspace/a.txt"})
    assert decision.allowed is False
    assert any("default_action=deny" in reason for reason in decision.reasons)


def test_default_action_deny_allows_matching_capability() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/*"]}}],
        "rules": [],
    }
    decision = _decision(policy, "read_file", {"path": "/workspace/a.txt"})
    assert decision.allowed is True


def test_path_glob_matching_workspace_csv() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/*.csv"]}}],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/report.csv"}).allowed is True
    assert _decision(policy, "read_file", {"path": "/workspace/report.json"}).allowed is False


def test_domain_glob_matching() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "web_search", "allow": {"domains": ["*.github.com"]}}],
        "rules": [],
    }
    assert _decision(policy, "web_search", {"url": "https://api.github.com/repos/x"}).allowed is True
    assert _decision(policy, "web_search", {"url": "https://example.com"}).allowed is False


def test_command_allowlist_exact_match() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "shell_execute", "allow": {"commands": ["git status", "git log", "ls"]}}],
        "rules": [],
    }
    assert _decision(policy, "shell_execute", {"command": "git status"}).allowed is True
    assert _decision(policy, "shell_execute", {"command": "git push"}).allowed is False


def test_wildcard_tool_applies_to_all_tools() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "*", "allow": {"paths": ["/workspace/*"]}}],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/a.txt"}).allowed is True
    assert _decision(policy, "write_file", {"path": "/workspace/b.txt"}).allowed is True


def test_capability_deny_blocks_specific_path() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [
            {"tool": "read_file", "allow": {"paths": ["/workspace/*"]}},
            {"tool": "*", "deny": {"paths": ["/workspace/secret/*"]}},
        ],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/public/a.txt"}).allowed is True
    denied = _decision(policy, "read_file", {"path": "/workspace/secret/token.txt"})
    assert denied.allowed is False
    assert any("denied by path constraint" in reason for reason in denied.reasons)


def test_backward_compat_no_capabilities_uses_existing_behavior() -> None:
    policy = {
        "rules": [],
        "tool_access": {"mode": "denylist", "denied": ["shell_execute"]},
    }
    assert _decision(policy, "read_file", {"path": "/tmp/a"}).allowed is True
    assert _decision(policy, "shell_execute", {"command": "id"}).allowed is False


def test_backward_compat_default_action_allow_uses_existing_denylist() -> None:
    policy = {
        "default_action": "allow",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/*"]}}],
        "tool_access": {"mode": "denylist", "denied": ["shell_execute"]},
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/etc/passwd"}).allowed is True
    assert _decision(policy, "shell_execute", {"command": "id"}).allowed is False


def test_mixed_mode_capabilities_and_denied_tools_coexist() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "shell_execute", "allow": {"commands": ["git status"]}}],
        "tool_access": {"mode": "denylist", "denied": ["shell_execute"]},
        "rules": [],
    }
    decision = _decision(policy, "shell_execute", {"command": "git status"})
    assert decision.allowed is False
    assert any("denylist" in reason for reason in decision.reasons)


def test_empty_capabilities_with_deny_blocks_everything() -> None:
    policy = {"default_action": "deny", "capabilities": [], "rules": []}
    assert _decision(policy, "web_search", {"url": "https://example.com"}).allowed is False
    assert _decision(policy, "read_file", {"path": "/tmp/a"}).allowed is False


def test_multiple_capabilities_for_same_tool_union_permissions() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [
            {"tool": "read_file", "allow": {"paths": ["/workspace/*"]}},
            {"tool": "read_file", "allow": {"paths": ["/tmp/*"]}},
        ],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/a.txt"}).allowed is True
    assert _decision(policy, "read_file", {"path": "/tmp/b.txt"}).allowed is True
    assert _decision(policy, "read_file", {"path": "/etc/passwd"}).allowed is False


def test_nested_path_glob_matches_data_json() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/data/**/*.json"]}}],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/data/x/y/z.json"}).allowed is True


def test_tool_names_are_case_insensitive() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/*"]}}],
        "rules": [],
    }
    assert _decision(policy, "READ_FILE", {"path": "/workspace/a.txt"}).allowed is True


def test_missing_path_param_does_not_fail_path_allow_constraint() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/*"]}}],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"mode": "metadata_only"}).allowed is True


def test_missing_domain_param_does_not_fail_domain_allow_constraint() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "web_search", "allow": {"domains": ["*.github.com"]}}],
        "rules": [],
    }
    assert _decision(policy, "web_search", {"query": "release notes"}).allowed is True


def test_missing_command_param_does_not_fail_command_allow_constraint() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "shell_execute", "allow": {"commands": ["git status"]}}],
        "rules": [],
    }
    assert _decision(policy, "shell_execute", {"cwd": "/workspace"}).allowed is True


def test_capability_deny_domain_blocks_even_when_allow_exists() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [
            {"tool": "web_search", "allow": {"domains": ["*.github.com", "*.google.com"]}},
            {"tool": "web_search", "deny": {"domains": ["*.google.com"]}},
        ],
        "rules": [],
    }
    assert _decision(policy, "web_search", {"url": "https://api.github.com"}).allowed is True
    assert _decision(policy, "web_search", {"url": "https://mail.google.com"}).allowed is False


def test_capability_deny_command_blocks() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [
            {"tool": "shell_execute", "allow": {"commands": ["git *"]}},
            {"tool": "shell_execute", "deny": {"commands": ["git push*"]}},
        ],
        "rules": [],
    }
    assert _decision(policy, "shell_execute", {"command": "git status"}).allowed is True
    assert _decision(policy, "shell_execute", {"command": "git push origin main"}).allowed is False


def test_capability_wildcard_allows_by_domain_for_any_tool() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "*", "allow": {"domains": ["*.github.com"]}}],
        "rules": [],
    }
    assert _decision(policy, "web_fetch", {"url": "https://api.github.com/repos"}).allowed is True
    assert _decision(policy, "web_fetch", {"url": "https://example.com"}).allowed is False


def test_capability_path_matching_uses_normalized_path() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/*"]}}],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/../workspace/a.txt"}).allowed is True


def test_capability_supports_filepath_key() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/tmp/*"]}}],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"filePath": "/tmp/a.txt"}).allowed is True


def test_default_action_allow_with_capability_deny_only_still_denies() -> None:
    policy = {
        "default_action": "allow",
        "capabilities": [{"tool": "*", "deny": {"paths": ["/etc/*"]}}],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/etc/passwd"}).allowed is False
    assert _decision(policy, "read_file", {"path": "/tmp/a"}).allowed is True


def test_capability_reason_when_constraints_not_met() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/*"]}}],
        "rules": [],
    }
    decision = _decision(policy, "read_file", {"path": "/root/id_rsa"})
    assert decision.allowed is False
    assert any("does not satisfy allow constraints" in reason for reason in decision.reasons)


def test_capability_and_rule_based_denied_paths_both_apply() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "read_file", "allow": {"paths": ["/workspace/*", "/etc/*"]}}],
        "rules": [{"name": "file_access", "denied_paths": ["/etc"]}],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/a.txt"}).allowed is True
    assert _decision(policy, "read_file", {"path": "/etc/passwd"}).allowed is False


def test_capability_can_allow_all_tool_params_when_allow_empty_mapping() -> None:
    policy = {"default_action": "deny", "capabilities": [{"tool": "web_search", "allow": {}}], "rules": []}
    assert _decision(policy, "web_search", {"query": "x", "url": "https://example.com"}).allowed is True


def test_capability_tool_wildcard_with_only_deny_blocks_all_tools() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [
            {"tool": "*", "allow": {"paths": ["/workspace/*"]}},
            {"tool": "*", "deny": {"paths": ["/workspace/secret/*"]}},
        ],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/secret/a.txt"}).allowed is False
    assert _decision(policy, "write_file", {"path": "/workspace/secret/b.txt"}).allowed is False


def test_capability_tool_mismatch_deny_default_blocks() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "web_search", "allow": {"domains": ["*.github.com"]}}],
        "rules": [],
    }
    assert _decision(policy, "read_file", {"path": "/workspace/a.txt"}).allowed is False


def test_capability_domain_can_be_plain_host() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "web_search", "allow": {"domains": ["api.github.com"]}}],
        "rules": [],
    }
    assert _decision(policy, "web_search", {"domain": "api.github.com"}).allowed is True


def test_capability_command_glob_supported() -> None:
    policy = {
        "default_action": "deny",
        "capabilities": [{"tool": "shell_execute", "allow": {"commands": ["git *", "ls*"]}}],
        "rules": [],
    }
    assert _decision(policy, "shell_execute", {"command": "git log --oneline"}).allowed is True
    assert _decision(policy, "shell_execute", {"command": "python -V"}).allowed is False


def test_load_policy_invalid_capability_structure_graceful(tmp_path: Path) -> None:
    path = tmp_path / "bad.yaml"
    path.write_text(
        """
default_action: deny
capabilities:
  - tool: read_file
    allow:
      paths: "/workspace/*"
rules: []
""".strip(),
        encoding="utf-8",
    )
    with pytest.raises(PolicyError):
        load_policy(path)


def test_load_policy_empty_capabilities_with_deny_is_valid(tmp_path: Path) -> None:
    path = tmp_path / "deny-all.yaml"
    path.write_text("default_action: deny\ncapabilities: []\nrules: []\n", encoding="utf-8")
    policy = load_policy(path)
    decision = _decision(policy, "read_file", {"path": "/workspace/a"})
    assert decision.allowed is False




from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone

from orchesis.contrib.ip_allowlist import IPAllowlistHandler
from orchesis.contrib.pii_detector import PIIDetectorHandler
from orchesis.contrib.time_window import TimeWindowHandler
from orchesis.engine import evaluate
from orchesis.plugins import PluginInfo, PluginRegistry, load_plugins_for_policy
from orchesis.state import RateLimitTracker
from orchesis.telemetry import InMemoryEmitter


@dataclass
class _SimpleHandler:
    reason: str = ""

    def evaluate(self, rule, request, **kwargs):  # noqa: ANN001, ANN003
        _ = (rule, request, kwargs)
        return ([self.reason] if self.reason else []), ["custom_rule"]


def test_register_plugin() -> None:
    registry = PluginRegistry()
    info = PluginInfo("test", "custom_rule", "1.0", "test plugin", _SimpleHandler())
    registry.register(info)
    assert registry.is_registered("custom_rule") is True


def test_cannot_override_builtin_type() -> None:
    registry = PluginRegistry()
    info = PluginInfo("bad", "file_access", "1.0", "bad", _SimpleHandler())
    try:
        registry.register(info)
        assert False, "Expected ValueError"
    except ValueError:
        assert True


def test_plugin_evaluation() -> None:
    registry = PluginRegistry()
    registry.register(
        PluginInfo(
            "custom", "custom_rule", "1.0", "custom", _SimpleHandler("custom_rule: blocked")
        )
    )
    policy = {"rules": [{"name": "r1", "type": "custom_rule"}]}
    decision = evaluate({"tool": "read_file", "params": {}, "cost": 0.0}, policy, plugins=registry)
    assert decision.allowed is False
    assert "custom_rule: blocked" in decision.reasons


def test_plugin_fail_closed() -> None:
    class Broken:
        def evaluate(self, rule, request, **kwargs):  # noqa: ANN001, ANN003
            _ = (rule, request, kwargs)
            raise RuntimeError("boom")

    registry = PluginRegistry()
    registry.register(PluginInfo("broken", "broken_rule", "1.0", "broken", Broken()))
    policy = {"rules": [{"name": "x", "type": "broken_rule"}]}
    decision = evaluate({"tool": "read_file", "params": {}, "cost": 0.0}, policy, plugins=registry)
    assert decision.allowed is False
    assert any("internal_error" in reason for reason in decision.reasons)


def test_plugin_in_debug_trace() -> None:
    registry = PluginRegistry()
    registry.register(PluginInfo("custom", "custom_rule", "1.0", "custom", _SimpleHandler()))
    decision = evaluate(
        {"tool": "read_file", "params": {}, "cost": 0.0},
        {"rules": [{"name": "x", "type": "custom_rule"}]},
        plugins=registry,
        debug=True,
    )
    assert isinstance(decision.debug_trace, dict)
    assert any(item["rule"] == "custom_rule" for item in decision.debug_trace["rule_results"])


def test_plugin_in_telemetry() -> None:
    registry = PluginRegistry()
    registry.register(PluginInfo("custom", "custom_rule", "1.0", "custom", _SimpleHandler()))
    emitter = InMemoryEmitter()
    _ = evaluate(
        {"tool": "read_file", "params": {}, "cost": 0.0},
        {"rules": [{"name": "x", "type": "custom_rule"}]},
        plugins=registry,
        emitter=emitter,
        state=RateLimitTracker(persist_path=None),
    )
    event = emitter.get_events()[0]
    assert "custom_rule" in event.rules_checked


def test_unregister_plugin() -> None:
    registry = PluginRegistry()
    registry.register(PluginInfo("custom", "custom_rule", "1.0", "custom", _SimpleHandler()))
    registry.unregister("custom_rule")
    assert registry.is_registered("custom_rule") is False


def test_pii_detector_blocks_email() -> None:
    handler = PIIDetectorHandler()
    reasons, _ = handler.evaluate(
        {"type": "pii_detector"},
        {"params": {"query": "contact me at user@example.com"}},
        state=RateLimitTracker(persist_path=None),
        agent_id="a",
        session_id="s",
    )
    assert any("email" in reason for reason in reasons)


def test_pii_detector_allows_clean() -> None:
    handler = PIIDetectorHandler()
    reasons, _ = handler.evaluate(
        {"type": "pii_detector"},
        {"params": {"query": "no sensitive content here"}},
        state=RateLimitTracker(persist_path=None),
        agent_id="a",
        session_id="s",
    )
    assert reasons == []


def test_time_window_allows_in_hours() -> None:
    handler = TimeWindowHandler()
    handler._now = lambda tz: datetime(2026, 2, 28, 10, 0, tzinfo=timezone.utc)  # type: ignore[method-assign]
    reasons, _ = handler.evaluate(
        {
            "type": "time_window",
            "allowed_hours": {"start": "09:00", "end": "17:00", "timezone": "UTC"},
        },
        {"params": {}},
        state=RateLimitTracker(persist_path=None),
        agent_id="a",
        session_id="s",
    )
    assert reasons == []


def test_time_window_blocks_outside() -> None:
    handler = TimeWindowHandler()
    handler._now = lambda tz: datetime(2026, 2, 28, 22, 0, tzinfo=timezone.utc)  # type: ignore[method-assign]
    reasons, _ = handler.evaluate(
        {
            "type": "time_window",
            "allowed_hours": {"start": "09:00", "end": "17:00", "timezone": "UTC"},
        },
        {"params": {}},
        state=RateLimitTracker(persist_path=None),
        agent_id="a",
        session_id="s",
    )
    assert any("outside allowed window" in reason for reason in reasons)


def test_ip_allowlist_allows_valid() -> None:
    handler = IPAllowlistHandler()
    reasons, _ = handler.evaluate(
        {"type": "ip_allowlist", "allowed_ips": ["10.0.0.0/8"]},
        {"context": {"source_ip": "10.1.2.3"}},
        state=RateLimitTracker(persist_path=None),
        agent_id="a",
        session_id="s",
    )
    assert reasons == []


def test_ip_allowlist_blocks_invalid() -> None:
    handler = IPAllowlistHandler()
    reasons, _ = handler.evaluate(
        {"type": "ip_allowlist", "allowed_ips": ["10.0.0.0/8"]},
        {"context": {"source_ip": "8.8.8.8"}},
        state=RateLimitTracker(persist_path=None),
        agent_id="a",
        session_id="s",
    )
    assert any("not in allowlist" in reason for reason in reasons)


def test_auto_discover_plugins_from_policy() -> None:
    policy = {"rules": [{"name": "x", "type": "pii_detector"}]}
    registry = load_plugins_for_policy(policy)
    assert registry.is_registered("pii_detector") is True


def test_multiple_plugins_evaluate_in_order() -> None:
    class First:
        def evaluate(self, rule, request, **kwargs):  # noqa: ANN001, ANN003
            _ = (rule, request, kwargs)
            return ["first: denied"], ["first"]

    class Second:
        def evaluate(self, rule, request, **kwargs):  # noqa: ANN001, ANN003
            _ = (rule, request, kwargs)
            return ["second: denied"], ["second"]

    registry = PluginRegistry()
    registry.register(PluginInfo("first", "first_rule", "1.0", "first", First()))
    registry.register(PluginInfo("second", "second_rule", "1.0", "second", Second()))
    policy = {
        "rules": [
            {"name": "a", "type": "first_rule"},
            {"name": "b", "type": "second_rule"},
        ]
    }
    decision = evaluate({"tool": "read_file", "params": {}, "cost": 0.0}, policy, plugins=registry)
    assert "first: denied" in decision.reasons
    assert "second: denied" in decision.reasons
    assert "first" in decision.rules_checked
    assert "second" in decision.rules_checked



from __future__ import annotations

import json

from orchesis.otel import OTelEmitter, SpanData, TraceContext
from orchesis.telemetry import DecisionEvent


def _event() -> DecisionEvent:
    return DecisionEvent(
        event_id="evt-1",
        timestamp="2026-01-01T00:00:00Z",
        agent_id="cursor",
        tool="read_file",
        params_hash="abc",
        cost=0.1,
        decision="DENY",
        reasons=["file_access: denied"],
        rules_checked=["budget_limit", "rate_limit", "file_access"],
        rules_triggered=["file_access"],
        evaluation_order=["identity_check", "budget_limit", "rate_limit", "file_access"],
        evaluation_duration_us=50,
        policy_version="pv-1",
        state_snapshot={
            "trace_id": "0123456789abcdef0123456789abcdef",
            "parent_span_id": "0123456789abcdef",
        },
    )


def test_span_data_has_required_fields() -> None:
    span = SpanData(
        trace_id="0123456789abcdef0123456789abcdef",
        span_id="0123456789abcdef",
        parent_span_id=None,
        operation="orchesis.evaluate",
        start_time_ns=1,
        end_time_ns=2,
        status="OK",
        attributes={},
        events=[],
    )
    assert len(span.trace_id) == 32
    assert len(span.span_id) == 16
    assert span.operation == "orchesis.evaluate"


def test_decision_to_span_mapping(tmp_path) -> None:
    emitter = OTelEmitter(path=str(tmp_path / "traces.jsonl"))
    span = emitter._decision_to_span(_event())
    assert span.attributes["orchesis.agent_id"] == "cursor"
    assert span.attributes["orchesis.tool"] == "read_file"
    assert span.attributes["orchesis.decision"] == "DENY"
    assert span.attributes["orchesis.policy_version"] == "pv-1"
    assert span.attributes["orchesis.evaluation_duration_us"] == 50


def test_otel_emitter_writes_jsonl(tmp_path) -> None:
    path = tmp_path / "traces.jsonl"
    emitter = OTelEmitter(path=str(path))
    emitter.emit(_event())
    line = path.read_text(encoding="utf-8").splitlines()[0]
    payload = json.loads(line)
    assert payload["operation"] == "orchesis.evaluate"
    assert payload["attributes"]["orchesis.tool"] == "read_file"


def test_span_events_per_rule(tmp_path) -> None:
    emitter = OTelEmitter(path=str(tmp_path / "traces.jsonl"))
    span = emitter._decision_to_span(_event())
    assert len(span.events) == 3
    assert all(item["name"] == "rule.checked" for item in span.events)


def test_trace_context_from_headers() -> None:
    headers = {"traceparent": "00-0123456789abcdef0123456789abcdef-0123456789abcdef-01"}
    context = TraceContext.from_headers(headers)
    assert context.trace_id == "0123456789abcdef0123456789abcdef"
    assert context.parent_span_id == "0123456789abcdef"


def test_trace_context_to_headers() -> None:
    context = TraceContext(
        trace_id="0123456789abcdef0123456789abcdef", parent_span_id="0123456789abcdef"
    )
    headers = context.to_headers()
    assert "traceparent" in headers
    assert headers["traceparent"].startswith("00-0123456789abcdef0123456789abcdef-")


def test_trace_context_generation() -> None:
    context = TraceContext.from_headers({})
    assert len(context.trace_id) == 32



from __future__ import annotations

import os
import tempfile
from typing import Any


def _load_pii_detector() -> Any:
    try:
        from orchesis.pii_detector import PiiDetector  # type: ignore
    except Exception:
        from orchesis.contrib.pii_detector import PiiDetector
    return PiiDetector


def _load_secret_scanner() -> Any:
    try:
        from orchesis.secret_scanner import SecretScanner  # type: ignore
    except Exception:
        from orchesis.contrib.secret_scanner import SecretScanner
    return SecretScanner


def _load_policy_loader() -> Any:
    from orchesis.config import load_policy

    return load_policy


PII_CRASH_INPUT = (
    b"&C+8\xe6\xbc\xa2\xe5\xad\x97\xe2\xad\x83\xe3\xa4\xb8\xe1\x84\x85\xee\xa8\xb9"
    b"\xea\xb6\xb6\xeb\x9b\xaa +1 202-555-0147\xd8\xa7\xd9\x84\xd8\xb9\xd8\xb1\xd8\xa8"
    b"\xd9\x8a\xd8\xa9998\xd1999\xb79+9999998\xb783+99919,\xb999+9\xb7\xc7\xcb\xc6\xc655H"
)

SECRET_CRASH_INPUT = b"\x00\x00\xff\xff\x8c,u,\x00\x00\x88"
PII_CRASH_CONTROL = b"2l" + (b"\x13" * 108)
SECRET_CRASH_V2 = b"\xec\xff\xff\xff\xff\xff\xec\x00\x00\x88"
POLICY_CRASH = b"#allow<<:\n-\n\xa02[:\n\n<<:\n-\n\xa02:\n<<:\n- 0b_:\n<*als<:\n\n"
PII_CRASH_R3 = b"\xe6\xb0\xb1\xe6\x85\x80\xe6\xbc\xa3\xe2\xbc\x88 123-45-6789\xad\xad\xad\xad\xad\xad\xad\xad\xb5\xad\xad\xad\xad\xad\xad\xad\xaf3\xad3|\x180"
SECRET_CRASH_R3 = b"P%\x14\x00\x00\x88"
POLICY_CRASH_R3 = b'cs: ["w\xdc\xdc,\xdcu\xc3\xc3\xc3\xc1\xdcU\xc3\xc3\xc39999999999999999999999999999999999999999999999999\xc3cao\xdcu\xc3\xc3\xc3\xc3\xdcu\xc3\xc3\xc3\xc3\xdcu\xc3\xc3\xc3\xc3&i'
PII_CRASH_R4 = b"z\x03@oj2x@e.ac x.cc s@g.acoj2x@e.ac x.cc s@g.ac r@o.ca@o@e.ac\xdfx7@e.ac \xef@x.c@x.cc s@g.c\xf3@g.8z \xad~\xad\xad\xad\xad\xad\xadYYx r@o.ca@o@e.ac\xdfx7@e.ac \xef@x.c@x.cc s@g.c\xf3@g.8z \xad~\xad\xad\xad\xad\xad\xadYYx"
SECRET_CRASH_R4 = b"\x1d\x02\x00\x00\x00"
PII_CRASH_R5 = b">+hhhhhhhhhhhhhhhhhh9+499999;99999\xb79+439hhhhhh(hhhhhhh$hhh4+499999;99999\xb79+439\xad\xad\xad123-5-\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff4311041440~85288"
SECRET_CRASH_R5 = b"5\xec\x00\x00\x88"
POLICY_CRASH_R5 = b"sess: at\nrul: !!int\nrules:\n  - a\xa1\n  - name: file_access\ne:"
PII_CRASH_R6 = b"\x00\x00I\xec\x9b\x86\xe3\x83\x86\xeb\xb0\x80\xe9\xb8\x80\xe3\x90\xb9\x00\x00I\xec\x9b\x86\xe3\x83\x86\xeb\xb0\x80\xe9\xb8\x80\xe3\x90\xb9\xe3\xa4\xb3\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xe3\xa4\xb3\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef\xbf\xbf\xef8088473695\xff\xff\xff\x100"
SECRET_CRASH_R6 = b"\x10%\x02\x00\x00\x88"


def test_pii_detector_fuzz_crash_regression() -> None:
    PiiDetector = _load_pii_detector()
    detector = PiiDetector(use_fast_matching=True)
    text = PII_CRASH_INPUT.decode("utf-8", errors="replace")
    findings = detector.scan_text(text)
    assert isinstance(findings, list)


def test_pii_detector_fuzz_crash_bytes_input() -> None:
    PiiDetector = _load_pii_detector()
    detector = PiiDetector(use_fast_matching=True)
    findings = detector.scan_text(PII_CRASH_INPUT)
    assert isinstance(findings, list)


def test_pii_detector_null_bytes() -> None:
    PiiDetector = _load_pii_detector()
    detector = PiiDetector(use_fast_matching=True)
    findings = detector.scan_text("\x00\x00\x00foo@example.com\x00")
    assert isinstance(findings, list)


def test_pii_detector_pure_binary() -> None:
    PiiDetector = _load_pii_detector()
    detector = PiiDetector(use_fast_matching=True)
    binary_blob = bytes(range(256))
    findings = detector.scan_text(binary_blob)
    assert isinstance(findings, list)


def test_secret_scanner_fuzz_crash_regression() -> None:
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner(use_fast_matching=True)
    text = SECRET_CRASH_INPUT.decode("utf-8", errors="replace")
    findings = scanner.scan_text(text)
    assert isinstance(findings, list)


def test_secret_scanner_fuzz_crash_bytes_input() -> None:
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner(use_fast_matching=True)
    findings = scanner.scan_text(SECRET_CRASH_INPUT)
    assert isinstance(findings, list)


def test_secret_scanner_null_bytes() -> None:
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner(use_fast_matching=True)
    findings = scanner.scan_text("\x00\x00\x00sk-proj-abcdefghijklmnopqrstuvwxyz1234567890\x00")
    assert isinstance(findings, list)


def test_secret_scanner_pure_binary() -> None:
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner(use_fast_matching=True)
    binary_blob = bytes(range(256))
    findings = scanner.scan_text(binary_blob)
    assert isinstance(findings, list)


def test_pii_detector_fuzz_crash_control_chars() -> None:
    PiiDetector = _load_pii_detector()
    detector = PiiDetector(use_fast_matching=True)
    text = PII_CRASH_CONTROL.decode("utf-8", errors="replace")
    findings = detector.scan_text(text)
    assert isinstance(findings, list)


def test_secret_scanner_fuzz_crash_invalid_utf8_v2() -> None:
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner(use_fast_matching=True)
    text = SECRET_CRASH_V2.decode("utf-8", errors="replace")
    findings = scanner.scan_text(text)
    assert isinstance(findings, list)


def test_policy_yaml_fuzz_crash_malformed_merge_keys() -> None:
    load_policy = _load_policy_loader()
    payload = POLICY_CRASH.decode("utf-8", errors="replace")
    policy_path = None
    try:
        with tempfile.NamedTemporaryFile(
            "w", suffix=".yaml", delete=False, encoding="utf-8"
        ) as tmp:
            tmp.write(payload)
            policy_path = tmp.name
        try:
            _ = load_policy(policy_path)
        except ValueError as exc:
            assert "Invalid YAML policy" in str(exc)
    finally:
        if isinstance(policy_path, str) and os.path.exists(policy_path):
            os.unlink(policy_path)


def test_policy_yaml_fuzz_crash_bytes_input() -> None:
    load_policy = _load_policy_loader()
    payload = POLICY_CRASH.decode("utf-8", errors="replace")
    policy_path = None
    try:
        with tempfile.NamedTemporaryFile(
            "w", suffix=".yaml", delete=False, encoding="utf-8"
        ) as tmp:
            tmp.write(payload)
            policy_path = tmp.name
        try:
            _ = load_policy(policy_path)
        except Exception:
            pass
    finally:
        if isinstance(policy_path, str) and os.path.exists(policy_path):
            os.unlink(policy_path)


def test_policy_yaml_null_and_nbsp() -> None:
    load_policy = _load_policy_loader()
    payload = "\x00\x00allow:\n\u00a0\u00a0- tool: test\n"
    policy_path = None
    try:
        with tempfile.NamedTemporaryFile(
            "w", suffix=".yaml", delete=False, encoding="utf-8"
        ) as tmp:
            tmp.write(payload)
            policy_path = tmp.name
        try:
            _ = load_policy(policy_path)
        except Exception:
            pass
    finally:
        if isinstance(policy_path, str) and os.path.exists(policy_path):
            os.unlink(policy_path)


def test_pii_detector_fuzz_crash_soft_hyphen() -> None:
    """PII detector must not crash on soft-hyphen sequences around SSN."""
    PiiDetector = _load_pii_detector()
    detector = PiiDetector()
    text = PII_CRASH_R3.decode("utf-8", errors="replace")
    result = detector.scan_text(text)
    assert isinstance(result, list)


def test_secret_scanner_fuzz_crash_minimal_binary() -> None:
    """Secret scanner must not crash on 6-byte binary with nulls."""
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner()
    text = SECRET_CRASH_R3.decode("utf-8", errors="replace")
    result = scanner.scan_text(text)
    assert isinstance(result, list)


def test_policy_yaml_fuzz_crash_long_digit_string() -> None:
    """Policy YAML parser must not crash on long digit sequences with invalid UTF-8."""
    load_policy = _load_policy_loader()
    text = POLICY_CRASH_R3.decode("utf-8", errors="replace")
    fd, path = tempfile.mkstemp(suffix=".yaml")
    try:
        os.write(fd, text.encode("utf-8", errors="replace"))
        os.close(fd)
        try:
            _ = load_policy(path)
        except Exception:
            pass
    finally:
        try:
            os.close(fd)
        except Exception:
            pass
        os.unlink(path)


def test_pii_detector_fuzz_crash_email_like_patterns() -> None:
    """PII detector must not crash on email-like patterns with invalid UTF-8 and soft hyphens."""
    PiiDetector = _load_pii_detector()
    detector = PiiDetector()
    text = PII_CRASH_R4.decode("utf-8", errors="replace")
    result = detector.scan_text(text)
    assert isinstance(result, list)


def test_secret_scanner_fuzz_crash_control_chars() -> None:
    """Secret scanner must not crash on 5-byte control char input with nulls."""
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner()
    text = SECRET_CRASH_R4.decode("utf-8", errors="replace")
    result = scanner.scan_text(text)
    assert isinstance(result, list)


def test_pii_detector_fuzz_crash_repeated_chars_with_ssn() -> None:
    """PII detector must not crash on repeated chars with partial SSN and 0xff bytes."""
    PiiDetector = _load_pii_detector()
    detector = PiiDetector()
    text = PII_CRASH_R5.decode("utf-8", errors="replace")
    result = detector.scan_text(text)
    assert isinstance(result, list)


def test_secret_scanner_fuzz_crash_short_binary_r5() -> None:
    """Secret scanner must not crash on 5-byte binary with 0xec prefix."""
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner()
    text = SECRET_CRASH_R5.decode("utf-8", errors="replace")
    result = scanner.scan_text(text)
    assert isinstance(result, list)


def test_policy_yaml_fuzz_crash_invalid_yaml_tags() -> None:
    """Policy YAML parser must not crash on !!int tag with invalid UTF-8."""
    load_policy = _load_policy_loader()
    text = POLICY_CRASH_R5.decode("utf-8", errors="replace")
    fd, path = tempfile.mkstemp(suffix=".yaml")
    try:
        os.write(fd, text.encode("utf-8", errors="replace"))
        os.close(fd)
        try:
            _ = load_policy(path)
        except Exception:
            pass
    finally:
        try:
            os.close(fd)
        except Exception:
            pass
        os.unlink(path)


def test_pii_detector_fuzz_crash_cjk_with_nulls_r6() -> None:
    """PII detector must not crash on CJK chars with null bytes and U+FFFF sequences."""
    PiiDetector = _load_pii_detector()
    detector = PiiDetector()
    text = PII_CRASH_R6.decode("utf-8", errors="replace")
    result = detector.scan_text(text)
    assert isinstance(result, list)


def test_secret_scanner_fuzz_crash_short_binary_r6() -> None:
    """Secret scanner must not crash on 6-byte binary with percent encoding byte."""
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner()
    text = SECRET_CRASH_R6.decode("utf-8", errors="replace")
    result = scanner.scan_text(text)
    assert isinstance(result, list)


def test_pii_guard_all_previous_crashes() -> None:
    PiiDetector = _load_pii_detector()
    detector = PiiDetector()
    payloads = [
        PII_CRASH_INPUT,
        PII_CRASH_CONTROL,
        PII_CRASH_R3,
        PII_CRASH_R4,
        PII_CRASH_R5,
        PII_CRASH_R6,
    ]
    for payload in payloads:
        text = payload.decode("utf-8", errors="replace")
        result = detector.scan_text(text)
        assert isinstance(result, list)


def test_secret_guard_all_previous_crashes() -> None:
    SecretScanner = _load_secret_scanner()
    scanner = SecretScanner()
    payloads = [
        SECRET_CRASH_INPUT,
        SECRET_CRASH_V2,
        SECRET_CRASH_R3,
        SECRET_CRASH_R4,
        SECRET_CRASH_R5,
        SECRET_CRASH_R6,
    ]
    for payload in payloads:
        text = payload.decode("utf-8", errors="replace")
        result = scanner.scan_text(text)
        assert isinstance(result, list)


def test_policy_guard_all_previous_crashes() -> None:
    load_policy = _load_policy_loader()
    payloads = [
        POLICY_CRASH,
        POLICY_CRASH_R3,
        POLICY_CRASH_R5,
        b"\x00\x00allow:\n - test",
        b"rules:\n - <<:\n   - bad:\n",
        b"not: [valid",
    ]
    for payload in payloads:
        text = payload.decode("utf-8", errors="replace")
        fd, path = tempfile.mkstemp(suffix=".yaml")
        try:
            os.write(fd, text.encode("utf-8", errors="replace"))
            os.close(fd)
            try:
                _ = load_policy(path)
            except Exception:
                pass
        finally:
            try:
                os.close(fd)
            except Exception:
                pass
            os.unlink(path)



"""Demo backend service for proxy integration tests."""

from typing import Any

from fastapi import FastAPI

app = FastAPI(title="Orchesis Demo Backend")


@app.get("/data")
def get_data() -> dict[str, list[str]]:
    """Return static demo data list."""
    return {"items": ["report.csv", "data.json"]}


@app.post("/execute")
def execute_action(payload: dict[str, Any]) -> dict[str, str]:
    """Accept action payload and return completion status."""
    _ = payload
    return {"status": "done"}


@app.delete("/files/{path:path}")
def delete_file(path: str) -> dict[str, bool]:
    """Simulate file delete endpoint."""
    _ = path
    return {"deleted": True}



from __future__ import annotations

import threading

from orchesis.ars import AgentReliabilityScore


def test_task_completion_wilson_score() -> None:
    ars = AgentReliabilityScore()
    for _ in range(8):
        ars.update("a", is_session_end=True, session_success=True)
    for _ in range(2):
        ars.update("a", is_session_end=True, session_success=False)
    result = ars.compute("a")
    assert result is not None
    assert 0 < result.components["task_completion"] < 100


def test_task_completion_no_data_neutral() -> None:
    ars = AgentReliabilityScore()
    ars.update("a")
    result = ars.compute("a")
    assert result is not None
    assert result.components["task_completion"] == 50.0


def test_task_completion_all_success() -> None:
    ars = AgentReliabilityScore()
    for _ in range(20):
        ars.update("a", is_session_end=True, session_success=True)
    result = ars.compute("a")
    assert result is not None
    assert result.components["task_completion"] > 80


def test_task_completion_all_failure() -> None:
    ars = AgentReliabilityScore()
    for _ in range(20):
        ars.update("a", is_session_end=True, session_success=False)
    result = ars.compute("a")
    assert result is not None
    assert result.components["task_completion"] < 10


def test_task_completion_small_sample_conservative() -> None:
    ars = AgentReliabilityScore()
    ars.update("a", is_session_end=True, session_success=True)
    result = ars.compute("a")
    assert result is not None
    assert result.components["task_completion"] < 90


def test_loop_freedom_no_loops() -> None:
    ars = AgentReliabilityScore()
    for _ in range(10):
        ars.update("a", loop_flagged=False)
    result = ars.compute("a")
    assert result is not None
    assert result.components["loop_freedom"] > 90


def test_loop_freedom_all_loops() -> None:
    ars = AgentReliabilityScore()
    for _ in range(10):
        ars.update("a", loop_flagged=True)
    result = ars.compute("a")
    assert result is not None
    assert result.components["loop_freedom"] < 20


def test_loop_freedom_laplace_smoothing() -> None:
    ars = AgentReliabilityScore()
    ars.update("a", loop_flagged=True)
    result = ars.compute("a")
    assert result is not None
    assert 0 < result.components["loop_freedom"] < 50


def test_cost_efficiency_under_budget() -> None:
    ars = AgentReliabilityScore()
    for _ in range(10):
        ars.update("a", cost_usd=1.0)
    result = ars.compute("a")
    assert result is not None
    assert result.components["cost_efficiency"] > 70


def test_cost_efficiency_at_budget() -> None:
    ars = AgentReliabilityScore()
    for _ in range(50):
        ars.update("a", cost_usd=1.0)
    result = ars.compute("a")
    assert result is not None
    assert result.components["cost_efficiency"] == 0.0


def test_cost_efficiency_over_budget() -> None:
    ars = AgentReliabilityScore()
    for _ in range(70):
        ars.update("a", cost_usd=1.0)
    result = ars.compute("a")
    assert result is not None
    assert result.components["cost_efficiency"] == 0.0


def test_latency_health_fast() -> None:
    ars = AgentReliabilityScore(latency_baseline_ms=2000)
    for _ in range(10):
        ars.update("a", latency_ms=100)
    result = ars.compute("a")
    assert result is not None
    assert result.components["latency_health"] == 100.0


def test_latency_health_slow() -> None:
    ars = AgentReliabilityScore(latency_baseline_ms=100)
    for _ in range(10):
        ars.update("a", latency_ms=1000)
    result = ars.compute("a")
    assert result is not None
    assert result.components["latency_health"] < 20


def test_latency_health_insufficient_data() -> None:
    ars = AgentReliabilityScore()
    for _ in range(3):
        ars.update("a", latency_ms=200)
    result = ars.compute("a")
    assert result is not None
    assert result.components["latency_health"] == 50.0


def test_context_stability_uniform() -> None:
    ars = AgentReliabilityScore()
    for _ in range(10):
        ars.update("a", token_count=100)
    result = ars.compute("a")
    assert result is not None
    assert result.components["context_stability"] == 100.0


def test_context_stability_high_variance() -> None:
    ars = AgentReliabilityScore()
    for value in [10, 500, 20, 600, 15, 700]:
        ars.update("a", token_count=value)
    result = ars.compute("a")
    assert result is not None
    assert result.components["context_stability"] < 40


def test_termination_quality_all_clean() -> None:
    ars = AgentReliabilityScore()
    for _ in range(10):
        ars.update("a", clean_termination=True)
    result = ars.compute("a")
    assert result is not None
    assert result.components["termination_quality"] == 100.0


def test_termination_quality_mixed() -> None:
    ars = AgentReliabilityScore()
    for _ in range(7):
        ars.update("a", clean_termination=True)
    for _ in range(3):
        ars.update("a", clean_termination=False)
    result = ars.compute("a")
    assert result is not None
    assert result.components["termination_quality"] == 70.0


def test_security_compliance_no_threats() -> None:
    ars = AgentReliabilityScore()
    for _ in range(10):
        ars.update("a", has_threat=False)
    result = ars.compute("a")
    assert result is not None
    assert result.components["security_compliance"] == 100.0


def test_security_compliance_all_threats() -> None:
    ars = AgentReliabilityScore()
    for _ in range(10):
        ars.update("a", has_threat=True)
    result = ars.compute("a")
    assert result is not None
    assert result.components["security_compliance"] == 0.0


def test_compute_perfect_agent() -> None:
    ars = AgentReliabilityScore()
    for _ in range(300):
        ars.update(
            "a",
            is_session_end=True,
            session_success=True,
            loop_flagged=False,
            cost_usd=0.01,
            latency_ms=50,
            token_count=100,
            clean_termination=True,
            has_threat=False,
        )
    result = ars.compute("a")
    assert result is not None
    assert result.score >= 90


def test_compute_terrible_agent() -> None:
    ars = AgentReliabilityScore()
    for _ in range(300):
        ars.update(
            "a",
            is_session_end=True,
            session_success=False,
            loop_flagged=True,
            cost_usd=5.0,
            latency_ms=8000,
            token_count=1000,
            clean_termination=False,
            has_threat=True,
        )
    result = ars.compute("a")
    assert result is not None
    assert result.score < 40


def test_compute_average_agent() -> None:
    ars = AgentReliabilityScore()
    for i in range(100):
        ars.update(
            "a",
            is_session_end=True,
            session_success=(i % 3 != 0),
            loop_flagged=(i % 10 == 0),
            cost_usd=0.3,
            latency_ms=600,
            token_count=120 + (i % 30),
            clean_termination=(i % 4 != 0),
            has_threat=(i % 6 == 0),
        )
    result = ars.compute("a")
    assert result is not None
    assert 40 <= result.score <= 90


def test_compute_weighted_correctly() -> None:
    weights = {
        "task_completion": 100,
        "loop_freedom": 0,
        "cost_efficiency": 0,
        "latency_health": 0,
        "context_stability": 0,
        "termination_quality": 0,
        "security_compliance": 0,
    }
    ars = AgentReliabilityScore(weights=weights)
    for _ in range(20):
        ars.update("a", is_session_end=True, session_success=True)
    result = ars.compute("a")
    assert result is not None
    assert result.score == result.components["task_completion"]


def test_grade_a() -> None:
    ars = AgentReliabilityScore()
    for _ in range(200):
        ars.update("a", is_session_end=True, session_success=True, latency_ms=100, token_count=100, clean_termination=True)
    result = ars.compute("a")
    assert result is not None
    assert result.grade in {"A", "B"}


def test_grade_b() -> None:
    ars = AgentReliabilityScore()
    for i in range(200):
        ars.update("a", is_session_end=True, session_success=(i % 5 != 0), latency_ms=300, token_count=120, clean_termination=True)
    result = ars.compute("a")
    assert result is not None
    assert result.grade in {"A", "B", "C"}


def test_grade_c() -> None:
    ars = AgentReliabilityScore()
    for i in range(80):
        ars.update("a", is_session_end=True, session_success=(i % 2 == 0), latency_ms=800, token_count=200, clean_termination=(i % 3 != 0))
    result = ars.compute("a")
    assert result is not None
    assert result.grade in {"B", "C", "D"}


def test_grade_d() -> None:
    ars = AgentReliabilityScore()
    for i in range(120):
        ars.update("a", is_session_end=True, session_success=(i % 4 == 0), loop_flagged=True, latency_ms=2000, clean_termination=False, has_threat=True)
    result = ars.compute("a")
    assert result is not None
    assert result.grade in {"D", "F"}


def test_grade_f() -> None:
    ars = AgentReliabilityScore()
    for _ in range(120):
        ars.update("a", is_session_end=True, session_success=False, loop_flagged=True, cost_usd=10, latency_ms=10000, token_count=5000, clean_termination=False, has_threat=True)
    result = ars.compute("a")
    assert result is not None
    assert result.grade == "F"


def test_confidence_low() -> None:
    ars = AgentReliabilityScore()
    for _ in range(10):
        ars.update("a")
    result = ars.compute("a")
    assert result is not None
    assert result.confidence == "low"


def test_confidence_medium() -> None:
    ars = AgentReliabilityScore()
    for _ in range(60):
        ars.update("a")
    result = ars.compute("a")
    assert result is not None
    assert result.confidence == "medium"


def test_confidence_high() -> None:
    ars = AgentReliabilityScore()
    for _ in range(220):
        ars.update("a")
    result = ars.compute("a")
    assert result is not None
    assert result.confidence == "high"


def test_compute_all_multiple_agents() -> None:
    ars = AgentReliabilityScore()
    for _ in range(20):
        ars.update("a", is_session_end=True, session_success=True)
    for _ in range(20):
        ars.update("b", is_session_end=True, session_success=False)
    results = ars.compute_all()
    assert len(results) == 2


def test_update_unknown_agent_creates() -> None:
    ars = AgentReliabilityScore()
    ars.update("new_agent")
    result = ars.compute("new_agent")
    assert result is not None


def test_disabled_returns_none() -> None:
    ars = AgentReliabilityScore(enabled=False)
    ars.update("a")
    assert ars.compute("a") is None


def test_empty_agent_id_ignored() -> None:
    ars = AgentReliabilityScore()
    ars.update("")
    assert ars.compute_all() == []


def test_concurrent_updates() -> None:
    ars = AgentReliabilityScore()

    def _worker() -> None:
        for _ in range(200):
            ars.update("a", latency_ms=200, token_count=100, has_threat=False)

    threads = [threading.Thread(target=_worker) for _ in range(8)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    result = ars.compute("a")
    assert result is not None
    assert result.sample_size == 1600


def test_stats_tracking() -> None:
    ars = AgentReliabilityScore()
    ars.update("a")
    ars.update("a")
    _ = ars.compute("a")
    stats = ars.stats
    assert stats["total_updates"] == 2
    assert stats["total_computes"] == 1
    assert stats["agents_tracked"] == 1



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
import gzip
import json
import os
from pathlib import Path
import socket
import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.request import Request as UrlRequest, urlopen

import pytest

from orchesis.diff import SessionDiffer
from orchesis.recorder import SessionRecord, SessionRecorder
from orchesis.replayer import ReplayConfig, SessionReplayer
from orchesis.proxy import HTTPProxyConfig, LLMHTTPProxy


def _record(session_id: str = "s1", idx: int = 0, status: int = 200, cost: float = 0.01) -> SessionRecord:
    return SessionRecord(
        request_id=f"r{idx}",
        session_id=session_id,
        timestamp=time.time() + idx,
        request={"model": "gpt-4o", "messages": [{"role": "user", "content": f"q{idx}"}]},
        response={"ok": True} if status < 400 else None,
        status_code=status,
        provider="openai",
        model="gpt-4o",
        latency_ms=10.0 + idx,
        cost=cost,
        error=None if status < 400 else "err",
        metadata={"agent_id": "a1"},
    )


def test_record_single_request_file_created(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record())
    files = list(tmp_path.glob("*.jsonl.gz"))
    assert len(files) == 1


def test_record_multiple_same_session_append(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 1))
    r.record(_record("s1", 2))
    loaded = r.load_session("s1")
    assert len(loaded) == 2


def test_record_different_sessions_different_files(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 1))
    r.record(_record("s2", 1))
    assert len(list(tmp_path.glob("*.jsonl.gz"))) == 2


def test_load_session_order(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 2))
    r.record(_record("s1", 1))
    loaded = r.load_session("s1")
    assert loaded[0].request_id == "r1"


def test_list_sessions_returns_summaries(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 1))
    r.close_session("s1")
    summaries = r.list_sessions()
    assert summaries and summaries[0].session_id == "s1"


def test_summary_stats_correct(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 1, status=200, cost=0.2))
    r.record(_record("s1", 2, status=500, cost=0.3))
    s = r.get_session_summary("s1")
    assert s.request_count == 2
    assert s.error_count == 1
    assert s.total_cost == pytest.approx(0.5)


def test_delete_session_removes_file(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 1))
    assert r.delete_session("s1") is True
    assert not list(tmp_path.glob("*.jsonl.gz"))


def test_cleanup_old_sessions(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 1))
    r.close_session("s1")
    path = list(tmp_path.glob("*.jsonl.gz"))[0]
    old = time.time() - (40 * 86400)
    os.utime(path, (old, old))
    deleted = r.cleanup(max_age_days=30)
    assert deleted >= 1


def test_compressed_file_valid_gzip(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path), compress=True)
    r.record(_record("s1", 1))
    file = list(tmp_path.glob("*.jsonl.gz"))[0]
    with gzip.open(file, "rt", encoding="utf-8") as fh:
        line = fh.readline()
    assert line.strip().startswith("{")


def test_thread_safety_concurrent_recording(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))

    def worker(i: int) -> None:
        for j in range(100):
            r.record(_record("s1", i * 1000 + j))

    with ThreadPoolExecutor(max_workers=10) as pool:
        for i in range(10):
            pool.submit(worker, i)
    assert len(r.load_session("s1")) == 1000


def test_file_rotation_at_record_limit(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path), max_records_per_file=2)
    for i in range(5):
        r.record(_record("s1", i))
    assert len(list(tmp_path.glob("*.jsonl.gz"))) >= 2


def test_missing_session_graceful(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    assert r.load_session("missing") == []


def test_corrupted_session_file_graceful(tmp_path: Path) -> None:
    p = tmp_path / "s1_20260101.jsonl.gz"
    p.write_bytes(b"notgzip")
    r = SessionRecorder(storage_path=str(tmp_path))
    assert r.load_session("s1") == []


def test_close_session_flushes(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 1))
    r.close_session("s1")
    assert len(r.load_session("s1")) == 1


def test_get_stats_fields(tmp_path: Path) -> None:
    r = SessionRecorder(storage_path=str(tmp_path))
    r.record(_record("s1", 1))
    stats = r.get_stats()
    assert "active_sessions" in stats and "total_recorded" in stats and "storage_size_bytes" in stats


def test_context_manager_closes(tmp_path: Path) -> None:
    with SessionRecorder(storage_path=str(tmp_path)) as r:
        r.record(_record("s1", 1))
    assert len(list(tmp_path.glob("*.jsonl.gz"))) == 1


def test_replayer_dry_run_model_override_no_http(tmp_path: Path, monkeypatch) -> None:
    session = [_record("s1", i) for i in range(3)]
    replayer = SessionReplayer()
    monkeypatch.setattr(replayer, "_send_replay", lambda body, path="/v1/chat/completions": (_ for _ in ()).throw(RuntimeError("no")))
    report = replayer.replay(session, ReplayConfig(model_override="gpt-4o-mini", dry_run=True))
    assert report.summary.total_requests == 3
    assert report.summary.replay_cost >= 0.0


def test_replayer_dry_run_with_policy_blocks() -> None:
    session = [_record("s1", 1)]
    replayer = SessionReplayer()
    policy = {"rules": [{"name": "sql_restriction", "denied_operations": ["DROP"]}]}
    session[0].request["tool"] = "sql_query"
    session[0].request["query"] = "DROP TABLE x"
    report = replayer.replay(session, ReplayConfig(dry_run=True), policy=policy)
    assert report.summary.total_requests == 1


def test_replay_report_totals() -> None:
    session = [_record("s1", i, cost=0.1) for i in range(4)]
    replayer = SessionReplayer()
    report = replayer.replay(session, ReplayConfig(dry_run=True))
    assert report.summary.original_cost == pytest.approx(0.4)


def test_cost_delta_calculation() -> None:
    session = [_record("s1", i, cost=0.1) for i in range(2)]
    r = SessionReplayer()
    report = r.replay(session, ReplayConfig(model_override="gpt-4o-mini", dry_run=True))
    assert isinstance(report.summary.cost_delta, float)


def test_max_requests_limit() -> None:
    session = [_record("s1", i) for i in range(10)]
    report = SessionReplayer().replay(session, ReplayConfig(dry_run=True, max_requests=3))
    assert len(report.results) == 3


def test_empty_session_replay() -> None:
    report = SessionReplayer().replay([], ReplayConfig(dry_run=True))
    assert report.summary.total_requests == 0


def test_estimate_cost_method() -> None:
    session = [_record("s1", i) for i in range(3)]
    est = SessionReplayer().estimate_cost(session, "gpt-4o-mini")
    assert est >= 0.0


def test_diff_identical_zero_delta() -> None:
    session = [_record("s1", 1, cost=0.2)]
    replay = SessionReplayer().replay(session, ReplayConfig(dry_run=True))
    diff = SessionDiffer().diff(session, replay)
    assert diff.cost_comparison.delta == pytest.approx(replay.summary.cost_delta)


def test_diff_cost_savings_percentage() -> None:
    session = [_record("s1", i, cost=0.5) for i in range(3)]
    replay = SessionReplayer().replay(session, ReplayConfig(model_override="gpt-4o-mini", dry_run=True))
    diff = SessionDiffer().diff(session, replay)
    assert isinstance(diff.cost_comparison.savings_pct, float)


def test_diff_new_errors_listed() -> None:
    session = [_record("s1", 1, status=200)]
    replay = SessionReplayer().replay(session, ReplayConfig(dry_run=True))
    replay.results[0].replay_status = 500
    replay.results[0].replay_error = "boom"
    diff = SessionDiffer().diff(session, replay)
    assert diff.error_comparison.new_errors


def test_diff_policy_impact_populated() -> None:
    session = [_record("s1", 1, cost=0.4)]
    replay = SessionReplayer().replay(session, ReplayConfig(dry_run=True))
    replay.results[0].policy_blocked = True
    replay.results[0].policy_block_reason = "blocked"
    diff = SessionDiffer().diff(session, replay)
    assert diff.policy_impact.blocked_count == 1


def test_diff_to_json_valid() -> None:
    session = [_record("s1", 1)]
    replay = SessionReplayer().replay(session, ReplayConfig(dry_run=True))
    diff = SessionDiffer().diff(session, replay)
    payload = SessionDiffer().to_json(diff)
    assert isinstance(json.loads(payload), dict)


def test_diff_to_summary_text_readable() -> None:
    session = [_record("s1", 1)]
    replay = SessionReplayer().replay(session, ReplayConfig(dry_run=True))
    diff = SessionDiffer().diff(session, replay)
    text = SessionDiffer().to_summary_text(diff)
    assert "Session:" in text


class _RecorderUpstreamHandler(BaseHTTPRequestHandler):
    def do_POST(self) -> None:  # noqa: N802
        length = int(self.headers.get("Content-Length", "0") or "0")
        _ = self.rfile.read(length)
        payload = {
            "model": "gpt-4o-mini",
            "usage": {"prompt_tokens": 5, "completion_tokens": 3},
            "choices": [{"message": {"content": "ok"}, "finish_reason": "stop"}],
        }
        data = json.dumps(payload).encode("utf-8")
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.send_header("Content-Length", str(len(data)))
        self.end_headers()
        self.wfile.write(data)

    def log_message(self, fmt: str, *args) -> None:
        _ = (fmt, args)


def _start_http_server(handler_cls: type[BaseHTTPRequestHandler]) -> tuple[HTTPServer, threading.Thread]:
    server = HTTPServer(("127.0.0.1", 0), handler_cls)
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    return server, thread


def _pick_port() -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    p = int(sock.getsockname()[1])
    sock.close()
    return p


def _post(port: int) -> dict[str, str]:
    req = UrlRequest(
        f"http://127.0.0.1:{port}/v1/chat/completions",
        data=json.dumps({"model": "gpt-4o", "messages": [{"role": "user", "content": "hi"}]}).encode("utf-8"),
        headers={"Content-Type": "application/json", "Authorization": "Bearer x"},
        method="POST",
    )
    with urlopen(req, timeout=5) as resp:
        _ = resp.read()
        return dict(resp.headers.items())


def test_proxy_recording_enabled_headers_and_file(tmp_path: Path) -> None:
    upstream, _ = _start_http_server(_RecorderUpstreamHandler)
    policy = tmp_path / "policy.yaml"
    policy.write_text(
        """
rules: []
recording:
  enabled: true
  storage_path: ".orchesis/sessions"
  max_age_days: 30
  max_file_size_mb: 10
  compress: true
  include_response_body: true
  exclude_models: []
""".strip(),
        encoding="utf-8",
    )
    port = _pick_port()
    cwd = os.getcwd()
    os.chdir(tmp_path)
    proxy = LLMHTTPProxy(
        policy_path=str(policy),
        config=HTTPProxyConfig(
            host="127.0.0.1",
            port=port,
            upstream={"openai": f"http://127.0.0.1:{upstream.server_address[1]}", "anthropic": f"http://127.0.0.1:{upstream.server_address[1]}"},
        ),
    )
    proxy.start(blocking=False)
    try:
        headers = _post(port)
        assert headers.get("X-Orchesis-Session-Id")
        assert headers.get("X-Orchesis-Request-Id")
        sessions_dir = tmp_path / ".orchesis" / "sessions"
        assert sessions_dir.exists()
        assert list(sessions_dir.glob("*.jsonl.gz"))
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()
        os.chdir(cwd)


def test_proxy_recording_disabled_no_session_headers(tmp_path: Path) -> None:
    upstream, _ = _start_http_server(_RecorderUpstreamHandler)
    policy = tmp_path / "policy.yaml"
    policy.write_text("rules: []\n", encoding="utf-8")
    port = _pick_port()
    proxy = LLMHTTPProxy(
        policy_path=str(policy),
        config=HTTPProxyConfig(
            host="127.0.0.1",
            port=port,
            upstream={"openai": f"http://127.0.0.1:{upstream.server_address[1]}", "anthropic": f"http://127.0.0.1:{upstream.server_address[1]}"},
        ),
    )
    proxy.start(blocking=False)
    try:
        headers = _post(port)
        assert "X-Orchesis-Session-Id" not in headers
        assert "X-Orchesis-Request-Id" not in headers
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_proxy_get_sessions_endpoint(tmp_path: Path) -> None:
    upstream, _ = _start_http_server(_RecorderUpstreamHandler)
    policy = tmp_path / "policy.yaml"
    policy.write_text("rules: []\nrecording:\n  enabled: true\n", encoding="utf-8")
    port = _pick_port()
    cwd = os.getcwd()
    os.chdir(tmp_path)
    proxy = LLMHTTPProxy(
        policy_path=str(policy),
        config=HTTPProxyConfig(
            host="127.0.0.1",
            port=port,
            upstream={"openai": f"http://127.0.0.1:{upstream.server_address[1]}", "anthropic": f"http://127.0.0.1:{upstream.server_address[1]}"},
        ),
    )
    proxy.start(blocking=False)
    try:
        _post(port)
        with urlopen(f"http://127.0.0.1:{port}/sessions", timeout=5) as resp:
            payload = json.loads(resp.read().decode("utf-8"))
        assert isinstance(payload.get("sessions"), list)
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()
        os.chdir(cwd)


def test_proxy_get_session_summary_endpoint(tmp_path: Path) -> None:
    upstream, _ = _start_http_server(_RecorderUpstreamHandler)
    policy = tmp_path / "policy.yaml"
    policy.write_text("rules: []\nrecording:\n  enabled: true\n", encoding="utf-8")
    port = _pick_port()
    cwd = os.getcwd()
    os.chdir(tmp_path)
    proxy = LLMHTTPProxy(
        policy_path=str(policy),
        config=HTTPProxyConfig(
            host="127.0.0.1",
            port=port,
            upstream={"openai": f"http://127.0.0.1:{upstream.server_address[1]}", "anthropic": f"http://127.0.0.1:{upstream.server_address[1]}"},
        ),
    )
    proxy.start(blocking=False)
    try:
        headers = _post(port)
        sid = headers["X-Orchesis-Session-Id"]
        with urlopen(f"http://127.0.0.1:{port}/sessions/{sid}", timeout=5) as resp:
            payload = json.loads(resp.read().decode("utf-8"))
        assert payload.get("session", {}).get("session_id") == sid
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()
        os.chdir(cwd)


def test_proxy_recorded_session_loadable(tmp_path: Path) -> None:
    recorder = SessionRecorder(storage_path=str(tmp_path))
    recorder.record(_record("s1", 1))
    loaded = recorder.load_session("s1")
    assert loaded and loaded[0].session_id == "s1"


def test_proxy_stats_include_recorder_section(tmp_path: Path) -> None:
    upstream, _ = _start_http_server(_RecorderUpstreamHandler)
    policy = tmp_path / "policy.yaml"
    policy.write_text("rules: []\nrecording:\n  enabled: true\n", encoding="utf-8")
    port = _pick_port()
    cwd = os.getcwd()
    os.chdir(tmp_path)
    proxy = LLMHTTPProxy(
        policy_path=str(policy),
        config=HTTPProxyConfig(
            host="127.0.0.1",
            port=port,
            upstream={"openai": f"http://127.0.0.1:{upstream.server_address[1]}", "anthropic": f"http://127.0.0.1:{upstream.server_address[1]}"},
        ),
    )
    proxy.start(blocking=False)
    try:
        _post(port)
        with urlopen(f"http://127.0.0.1:{port}/stats", timeout=5) as resp:
            payload = json.loads(resp.read().decode("utf-8"))
        assert "recorder" in payload
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()
        os.chdir(cwd)



from pathlib import Path

import pytest

from orchesis.config import PolicyError, load_policy, validate_policy


def _write_yaml(tmp_path: Path, name: str, content: str) -> Path:
    file_path = tmp_path / name
    file_path.write_text(content, encoding="utf-8")
    return file_path


def test_load_policy_reads_valid_yaml_mapping(tmp_path: Path) -> None:
    policy_path = _write_yaml(
        tmp_path,
        "policy.yaml",
        """
rules:
  - name: budget_limit
    max_cost_per_call: 0.5
""".strip(),
    )

    policy = load_policy(policy_path)

    assert isinstance(policy, dict)
    assert policy["rules"][0]["name"] == "budget_limit"


def test_load_policy_raises_on_non_mapping_root(tmp_path: Path) -> None:
    policy_path = _write_yaml(
        tmp_path,
        "invalid.yaml",
        """
- name: budget_limit
  max_cost_per_call: 0.5
""".strip(),
    )

    with pytest.raises(ValueError, match="top-level YAML object"):
        load_policy(policy_path)


def test_load_policy_raises_policy_error_for_fuzzed_non_mapping_yaml(tmp_path: Path) -> None:
    policy_path = _write_yaml(tmp_path, "fuzzed.yaml", "[0b_\n#2\n")
    with pytest.raises((PolicyError, ValueError)):
        load_policy(policy_path)


def test_load_policy_raises_policy_error_for_valid_non_mapping_yaml(tmp_path: Path) -> None:
    policy_path = _write_yaml(tmp_path, "list-root.yaml", "[0b1]")
    with pytest.raises(PolicyError, match="top-level YAML object"):
        load_policy(policy_path)


def test_validate_policy_accepts_valid_policy() -> None:
    policy = {
        "rules": [
            {"name": "budget_limit", "max_cost_per_call": 0.5, "daily_budget": 50.0},
            {"name": "file_access", "allowed_paths": ["/tmp"], "denied_paths": ["/etc"]},
            {
                "name": "sql_restriction",
                "allowed_operations": ["SELECT"],
                "denied_operations": ["DROP"],
            },
            {"name": "rate_limit", "max_requests_per_minute": 100},
        ]
    }

    errors = validate_policy(policy)

    assert errors == []


def test_validate_policy_requires_rules_list() -> None:
    errors = validate_policy({})
    assert "policy.rules must be a list" in errors


def test_validate_policy_requires_rule_name() -> None:
    policy = {"rules": [{"max_cost_per_call": 0.5}]}

    errors = validate_policy(policy)

    assert "rules[0].name must be a non-empty string" in errors


def test_validate_policy_checks_required_fields_per_known_rule() -> None:
    policy = {
        "rules": [
            {"name": "budget_limit"},
            {"name": "file_access"},
            {"name": "sql_restriction"},
            {"name": "rate_limit"},
        ]
    }

    errors = validate_policy(policy)

    assert "rules[0].max_cost_per_call is required for budget_limit" in errors
    assert "rules[1] must define allowed_paths and/or denied_paths for file_access" in errors
    assert "rules[2].denied_operations is required for sql_restriction" in errors
    assert "rules[3].max_requests_per_minute is required for rate_limit" in errors


def test_load_policy_normalizes_default_action_and_capabilities(tmp_path: Path) -> None:
    policy_path = _write_yaml(
        tmp_path,
        "caps.yaml",
        """
default_action: deny
capabilities:
  - tool: READ_FILE
    allow:
      paths:
        - /workspace/*
  - tool: "*"
    deny:
      domains:
        - "*.evil.com"
rules: []
""".strip(),
    )
    policy = load_policy(policy_path)
    assert policy["default_action"] == "deny"
    assert policy["capabilities"][0]["tool"] == "read_file"
    assert policy["capabilities"][1]["tool"] == "*"


def test_load_policy_defaults_to_allow_when_default_action_missing(tmp_path: Path) -> None:
    policy_path = _write_yaml(tmp_path, "default-action.yaml", "rules: []")
    policy = load_policy(policy_path)
    assert policy["default_action"] == "allow"
    assert policy["capabilities"] == []


def test_load_policy_raises_on_invalid_default_action(tmp_path: Path) -> None:
    policy_path = _write_yaml(
        tmp_path,
        "bad-default-action.yaml",
        """
default_action: block
rules: []
""".strip(),
    )
    with pytest.raises(PolicyError, match="default_action"):
        load_policy(policy_path)


def test_load_policy_raises_on_invalid_capability_structure(tmp_path: Path) -> None:
    policy_path = _write_yaml(
        tmp_path,
        "bad-capability.yaml",
        """
default_action: deny
capabilities:
  - tool: read_file
    allow: "/workspace/*"
rules: []
""".strip(),
    )
    with pytest.raises(PolicyError, match="capabilities\\[0\\]\\.allow"):
        load_policy(policy_path)


def test_load_policy_raises_on_capability_without_allow_or_deny(tmp_path: Path) -> None:
    policy_path = _write_yaml(
        tmp_path,
        "empty-capability.yaml",
        """
default_action: deny
capabilities:
  - tool: read_file
rules: []
""".strip(),
    )
    with pytest.raises(PolicyError, match="must define 'allow' and/or 'deny'"):
        load_policy(policy_path)



from __future__ import annotations

import json
import socket
import threading
import time
from urllib.error import HTTPError
from urllib.request import Request as UrlRequest, urlopen

from orchesis.demo import DEMO_STATS, DemoServer


def _pick_port() -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = int(sock.getsockname()[1])
    sock.close()
    return port


def _get_json(url: str) -> dict:
    with _urlopen_with_retry(url, timeout=5) as resp:
        return json.loads(resp.read().decode("utf-8"))


def _urlopen_with_retry(request_or_url, timeout: int = 5, attempts: int = 3):
    last_error = None
    for _ in range(max(1, attempts)):
        try:
            return urlopen(request_or_url, timeout=timeout)
        except Exception as error:
            last_error = error
            time.sleep(0.05)
    if last_error is not None:
        raise last_error
    return urlopen(request_or_url, timeout=timeout)


def _wait_for_server(port: int, attempts: int = 20) -> None:
    for _ in range(max(1, attempts)):
        try:
            with _urlopen_with_retry(f"http://127.0.0.1:{port}/dashboard", timeout=1, attempts=1):
                return
        except Exception:
            time.sleep(0.05)
    raise RuntimeError(f"demo server did not start on port {port}")


def test_demo_stats_valid_json() -> None:
    assert isinstance(DEMO_STATS, dict)
    assert "cost_velocity" in DEMO_STATS


def test_demo_server_starts() -> None:
    port = _pick_port()
    thread = threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True)
    thread.start()
    _wait_for_server(port)
    with _urlopen_with_retry(f"http://127.0.0.1:{port}/dashboard", timeout=5) as resp:
        assert int(resp.status) == 200


def test_demo_dashboard_has_banner() -> None:
    port = _pick_port()
    thread = threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True)
    thread.start()
    _wait_for_server(port)
    with _urlopen_with_retry(f"http://127.0.0.1:{port}/dashboard", timeout=5) as resp:
        body = resp.read().decode("utf-8")
    assert "DEMO MODE" in body


def test_demo_stats_endpoint() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    payload = _get_json(f"http://127.0.0.1:{port}/stats")
    assert "cost_today" in payload
    assert "cost_velocity" in payload


def test_demo_agents_endpoint() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    payload = _get_json(f"http://127.0.0.1:{port}/api/dashboard/agents")
    assert "agents" in payload
    assert isinstance(payload["agents"], list)


def test_demo_mode_no_real_proxy() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    req = UrlRequest(
        f"http://127.0.0.1:{port}/v1/chat/completions",
        data=b"{}",
        method="POST",
    )
    try:
        _urlopen_with_retry(req, timeout=5)
        assert False
    except HTTPError as error:
        assert error.code == 404


def test_demo_approvals_endpoint() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    payload = _get_json(f"http://127.0.0.1:{port}/api/v1/approvals")
    assert "pending" in payload
    assert "history" in payload


def test_demo_approve_action() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    req = UrlRequest(f"http://127.0.0.1:{port}/api/v1/approvals/demo-approval-1/approve", data=b"{}", method="POST")
    with _urlopen_with_retry(req, timeout=5) as resp:
        payload = json.loads(resp.read().decode("utf-8"))
    assert payload["approved"] is True


def test_demo_deny_action() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    req = UrlRequest(f"http://127.0.0.1:{port}/api/v1/approvals/demo-approval-2/deny", data=b"{}", method="POST")
    with _urlopen_with_retry(req, timeout=5) as resp:
        payload = json.loads(resp.read().decode("utf-8"))
    assert payload["denied"] is True


def test_demo_overview_contains_compliance() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    payload = _get_json(f"http://127.0.0.1:{port}/api/dashboard/overview")
    assert "compliance_overview" in payload




from __future__ import annotations

from orchesis.cost_reporter import CostReporter
from orchesis.cost_tracker import CostTracker
from orchesis.loop_detector import LoopDetector


def _build_reporter(with_loop: bool = True) -> CostReporter:
    tracker = CostTracker(tool_costs={"web_search": 0.5, "read_file": 0.1})
    tracker.record_call("web_search")
    tracker.record_call("read_file")
    loop = LoopDetector(warn_threshold=1, block_threshold=2) if with_loop else None
    if loop is not None:
        loop.check_request({"model": "test", "messages": [], "tool_calls": [{"name": "web_search"}], "content_text": "x"})
        loop.check_request({"model": "test", "messages": [], "tool_calls": [{"name": "web_search"}], "content_text": "x"})
    return CostReporter(tracker, loop_detector=loop)


def test_daily_summary_structure() -> None:
    reporter = _build_reporter()
    summary = reporter.daily_summary()
    assert "date" in summary
    assert "total_usd" in summary
    assert "top_tools" in summary
    assert "hourly_breakdown" in summary
    assert "total_calls" in summary


def test_console_format_contains_header() -> None:
    reporter = _build_reporter()
    text = reporter.format_console()
    assert "Orchesis Cost Report" in text
    assert "Top tools by cost" in text


def test_markdown_format_contains_table() -> None:
    reporter = _build_reporter()
    text = reporter.format_markdown()
    assert "| Tool | Cost |" in text
    assert "# Orchesis Cost Report" in text


def test_top_tools_sorted_by_cost() -> None:
    reporter = _build_reporter(with_loop=False)
    summary = reporter.daily_summary()
    top = summary["top_tools"]
    assert top[0]["cost_usd"] >= top[-1]["cost_usd"]


def test_hourly_breakdown_present() -> None:
    reporter = _build_reporter(with_loop=False)
    summary = reporter.daily_summary()
    assert isinstance(summary["hourly_breakdown"], dict)


def test_loop_stats_included_when_detector_present() -> None:
    reporter = _build_reporter(with_loop=True)
    summary = reporter.daily_summary()
    assert "loops" in summary
    assert "saved_by_loop_detection" in summary


def test_loop_stats_not_included_without_detector() -> None:
    reporter = _build_reporter(with_loop=False)
    summary = reporter.daily_summary()
    assert "loops" not in summary


def test_empty_data_handled_gracefully_console() -> None:
    reporter = CostReporter(CostTracker(), loop_detector=None)
    text = reporter.format_console()
    assert "(no calls)" in text


def test_empty_data_handled_gracefully_markdown() -> None:
    reporter = CostReporter(CostTracker(), loop_detector=None)
    text = reporter.format_markdown()
    assert "| (no calls) | $0.0000 |" in text


def test_daily_summary_total_matches_tracker() -> None:
    tracker = CostTracker(tool_costs={"x": 1.25})
    tracker.record_call("x")
    reporter = CostReporter(tracker)
    summary = reporter.daily_summary()
    assert summary["total_usd"] == 1.25


def test_console_includes_loop_lines_when_loops_exist() -> None:
    reporter = _build_reporter(with_loop=True)
    text = reporter.format_console()
    assert "Loop detection" in text
    assert "Saved by loop detection" in text


def test_markdown_includes_hourly_section_when_non_empty() -> None:
    reporter = _build_reporter()
    text = reporter.format_markdown()
    assert "## Hourly Breakdown" in text




"""Stdlib HTTP server for scanner APIs."""

from __future__ import annotations

import json
import logging
import signal
import tempfile
import threading
import time
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from typing import Any

from orchesis import __version__
from orchesis.compliance import FRAMEWORK_CHECKS
from orchesis.contrib.ioc_database import IoCMatcher
from orchesis.scanner import McpConfigScanner, PolicyScanner, ScanFinding, SkillScanner

LOGGER = logging.getLogger("orchesis.scanner_server")
MAX_BODY_BYTES = 10 * 1024 * 1024


def _risk_level(score: int) -> str:
    if score >= 80:
        return "HIGH"
    if score >= 50:
        return "MEDIUM"
    if score >= 20:
        return "LOW"
    return "INFO"


def _finding_to_payload(finding: ScanFinding) -> dict[str, str]:
    return {
        "severity": str(finding.severity).upper(),
        "check": str(finding.category),
        "message": str(finding.description),
    }


def _scan_content_with(scanner: Any, suffix: str, content: str) -> list[ScanFinding]:
    with tempfile.NamedTemporaryFile("w", encoding="utf-8", suffix=suffix, delete=False) as handle:
        temp_path = Path(handle.name)
        handle.write(content)
    try:
        report = scanner.scan(str(temp_path))
        return report.findings
    finally:
        try:
            temp_path.unlink(missing_ok=True)
        except Exception:
            pass


def create_scanner_http_server(
    *,
    host: str = "127.0.0.1",
    port: int = 8081,
    policy_path: str | None = None,
    allow_file_access: bool = False,
) -> ThreadingHTTPServer:
    started_at = time.monotonic()
    skill_scanner = SkillScanner()
    mcp_scanner = McpConfigScanner()
    policy_scanner = PolicyScanner()
    ioc_matcher = IoCMatcher()
    _ = policy_path

    class ScannerHandler(BaseHTTPRequestHandler):
        server_version = "OrchesisScannerAPI/1.0"

        def _send_json(self, status: int, payload: dict[str, Any]) -> None:
            body = json.dumps(payload, ensure_ascii=False).encode("utf-8")
            self.send_response(status)
            self.send_header("Content-Type", "application/json; charset=utf-8")
            self.send_header("Content-Length", str(len(body)))
            self.end_headers()
            self.wfile.write(body)

        def _error(self, status: int, message: str) -> None:
            self._send_json(status, {"status": "error", "message": message})

        def _read_json_body(self) -> dict[str, Any] | None:
            content_type = str(self.headers.get("Content-Type", ""))
            if "application/json" not in content_type.lower():
                self._error(415, "Content-Type must be application/json")
                return None
            length_header = self.headers.get("Content-Length")
            if length_header is None:
                self._error(400, "Missing Content-Length header")
                return None
            try:
                content_length = int(length_header)
            except ValueError:
                self._error(400, "Invalid Content-Length header")
                return None
            if content_length < 0:
                self._error(400, "Invalid Content-Length header")
                return None
            if content_length > MAX_BODY_BYTES:
                self._error(413, "Request body too large")
                return None
            body = self.rfile.read(content_length)
            if len(body) > MAX_BODY_BYTES:
                self._error(413, "Request body too large")
                return None
            try:
                payload = json.loads(body.decode("utf-8"))
            except Exception:
                self._error(400, "Invalid JSON body")
                return None
            if not isinstance(payload, dict):
                self._error(400, "JSON body must be an object")
                return None
            return payload

        def do_GET(self) -> None:  # noqa: N802
            started = time.perf_counter()
            status = 200
            try:
                if self.path == "/health":
                    self._send_json(
                        200,
                        {
                            "status": "healthy",
                            "version": __version__,
                            "uptime_seconds": int(max(0.0, time.monotonic() - started_at)),
                        },
                    )
                    return
                if self.path == "/frameworks":
                    self._send_json(200, {"frameworks": sorted(FRAMEWORK_CHECKS.keys())})
                    return
                if self.path in {"/scan/skill", "/scan/mcp", "/scan/policy", "/scan/ioc"}:
                    status = 405
                    self._error(405, "Method Not Allowed")
                    return
                status = 404
                self._error(404, "Not Found")
            finally:
                duration_ms = (time.perf_counter() - started) * 1000.0
                LOGGER.info(
                    "scanner_api request ts=%s method=%s path=%s status=%s duration_ms=%.2f",
                    time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
                    "GET",
                    self.path,
                    status,
                    duration_ms,
                )

        def do_POST(self) -> None:  # noqa: N802
            started = time.perf_counter()
            status = 200
            try:
                payload = self._read_json_body()
                if payload is None:
                    return
                if self.path == "/scan/skill":
                    content = payload.get("content")
                    if not isinstance(content, str):
                        status = 400
                        self._error(400, "Field 'content' must be a string")
                        return
                    findings = _scan_content_with(skill_scanner, ".md", content)
                    risk_score = min(100, sum({"CRITICAL": 40, "HIGH": 20, "MEDIUM": 10, "LOW": 5}.get(f.severity.upper(), 1) for f in findings))
                    self._send_json(
                        200,
                        {
                            "status": "ok",
                            "findings": [_finding_to_payload(item) for item in findings],
                            "risk_score": risk_score,
                            "risk_level": _risk_level(risk_score),
                        },
                    )
                    return
                if self.path == "/scan/mcp":
                    if "config" in payload:
                        config = payload.get("config")
                        if not isinstance(config, dict):
                            status = 400
                            self._error(400, "Field 'config' must be an object")
                            return
                        content = json.dumps(config, ensure_ascii=False)
                        findings = _scan_content_with(mcp_scanner, ".json", content)
                    elif "config_path" in payload:
                        config_path = payload.get("config_path")
                        if not isinstance(config_path, str) or not config_path.strip():
                            status = 400
                            self._error(400, "Field 'config_path' must be a string")
                            return
                        if not allow_file_access:
                            status = 403
                            self._error(403, "File access is disabled; use --allow-file-access")
                            return
                        source = Path(config_path)
                        if not source.exists() or not source.is_file():
                            status = 404
                            self._error(404, "config_path not found")
                            return
                        report = mcp_scanner.scan(str(source))
                        findings = report.findings
                    else:
                        status = 400
                        self._error(400, "Provide 'config' or 'config_path'")
                        return
                    risk_score = min(100, sum({"CRITICAL": 40, "HIGH": 20, "MEDIUM": 10, "LOW": 5}.get(f.severity.upper(), 1) for f in findings))
                    self._send_json(
                        200,
                        {
                            "status": "ok",
                            "findings": [_finding_to_payload(item) for item in findings],
                            "risk_score": risk_score,
                            "risk_level": _risk_level(risk_score),
                        },
                    )
                    return
                if self.path == "/scan/policy":
                    content = payload.get("content")
                    if not isinstance(content, str):
                        status = 400
                        self._error(400, "Field 'content' must be a string")
                        return
                    try:
                        findings = _scan_content_with(policy_scanner, ".yaml", content)
                    except Exception:
                        status = 400
                        self._error(400, "Invalid policy YAML")
                        return
                    score = max(0, 100 - min(100, sum({"CRITICAL": 40, "HIGH": 20, "MEDIUM": 10, "LOW": 5}.get(f.severity.upper(), 1) for f in findings)))
                    self._send_json(
                        200,
                        {
                            "status": "ok",
                            "findings": [_finding_to_payload(item) for item in findings],
                            "score": score,
                        },
                    )
                    return
                if self.path == "/scan/ioc":
                    content = payload.get("content")
                    if not isinstance(content, str):
                        status = 400
                        self._error(400, "Field 'content' must be a string")
                        return
                    self._send_json(200, {"status": "ok", "matches": ioc_matcher.scan_text(content)})
                    return
                status = 404
                self._error(404, "Not Found")
            finally:
                duration_ms = (time.perf_counter() - started) * 1000.0
                LOGGER.info(
                    "scanner_api request ts=%s method=%s path=%s status=%s duration_ms=%.2f",
                    time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
                    "POST",
                    self.path,
                    status,
                    duration_ms,
                )

        def do_PUT(self) -> None:  # noqa: N802
            self._error(405, "Method Not Allowed")

        def do_DELETE(self) -> None:  # noqa: N802
            self._error(405, "Method Not Allowed")

        def log_message(self, format: str, *args: Any) -> None:
            _ = (format, args)

    return ThreadingHTTPServer((host, int(port)), ScannerHandler)


def run_scanner_http_server(
    *,
    host: str = "127.0.0.1",
    port: int = 8081,
    policy_path: str | None = None,
    allow_file_access: bool = False,
) -> None:
    server = create_scanner_http_server(
        host=host,
        port=port,
        policy_path=policy_path,
        allow_file_access=allow_file_access,
    )
    stop_event = threading.Event()

    def _shutdown_handler(_signum: int, _frame: Any) -> None:
        if stop_event.is_set():
            return
        stop_event.set()
        server.shutdown()

    signal.signal(signal.SIGINT, _shutdown_handler)
    if hasattr(signal, "SIGTERM"):
        signal.signal(signal.SIGTERM, _shutdown_handler)
    try:
        server.serve_forever()
    finally:
        server.server_close()



"""PII detector and backward-compatible plugin handler."""

from __future__ import annotations

import re
import unicodedata
from copy import deepcopy
from typing import Any

from orchesis.fast_scanner import FastPIIDetector
from orchesis.input_guard import sanitize_text
from orchesis.plugins import PluginInfo

PII_PATTERNS: dict[str, tuple[re.Pattern[str], str, str]] = {
    "email": (
        re.compile(r"[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}"),
        "medium",
        "Email address",
    ),
    "phone_us": (
        re.compile(r"(?:\+1[-.\s]?)?\(?\d{3}\)?[-.\s]?\d{3}[-.\s]?\d{4}"),
        "medium",
        "US phone number",
    ),
    "phone_intl": (
        re.compile(r"\+\d{1,3}[-.\s]?\d{4,14}"),
        "medium",
        "International phone number",
    ),
    "ssn": (re.compile(r"\b\d{3}-\d{2}-\d{4}\b"), "critical", "US Social Security Number"),
    "passport_us": (re.compile(r"\b[A-Z]\d{8}\b"), "high", "Possible US passport number"),
    "credit_card_visa": (
        re.compile(r"\b4\d{3}[-\s]?\d{4}[-\s]?\d{4}[-\s]?\d{4}\b"),
        "critical",
        "Visa credit card number",
    ),
    "credit_card_mc": (
        re.compile(r"\b5[1-5]\d{2}[-\s]?\d{4}[-\s]?\d{4}[-\s]?\d{4}\b"),
        "critical",
        "Mastercard credit card number",
    ),
    "credit_card_amex": (
        re.compile(r"\b3[47]\d{2}[-\s]?\d{6}[-\s]?\d{5}\b"),
        "critical",
        "Amex credit card number",
    ),
    "iban": (
        re.compile(r"\b[A-Z]{2}\d{2}[A-Z0-9]{4}\d{7}([A-Z0-9]?\d{0,16})\b"),
        "high",
        "IBAN number",
    ),
    "swift_bic": (
        re.compile(r"\b[A-Z]{6}[A-Z0-9]{2}([A-Z0-9]{3})?\b"),
        "medium",
        "SWIFT/BIC code",
    ),
    "icd10_code": (re.compile(r"\b[A-Z]\d{2}(\.\d{1,4})?\b"), "low", "Possible ICD-10 code"),
    "npi_number": (re.compile(r"\b\d{10}\b"), "medium", "Possible NPI number"),
    "ipv4_address": (
        re.compile(r"\b(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\b"),
        "low",
        "IPv4 address",
    ),
    "mac_address": (
        re.compile(r"\b([0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}\b"),
        "low",
        "MAC address",
    ),
    "date_of_birth": (
        re.compile(r"(?i)\b(?:dob|date.of.birth|born)\s*[=:]\s*\d{1,2}[/\-]\d{1,2}[/\-]\d{2,4}"),
        "high",
        "Date of birth",
    ),
    "drivers_license": (
        re.compile(r"(?i)(?:driver'?s?\s*license|DL)\s*[#:]\s*[A-Z0-9]{5,15}"),
        "high",
        "Driver's license number",
    ),
}

CLASSIFICATION_LEVELS = {
    "public": 0,
    "internal": 1,
    "confidential": 2,
    "restricted": 3,
}

_SEVERITY_RANK = {"low": 0, "medium": 1, "high": 2, "critical": 3}
_CRITICAL_PATTERNS = {"ssn", "credit_card_visa", "credit_card_mc", "credit_card_amex"}


def preprocess_for_pii(text: str) -> list[str]:
    if not isinstance(text, str):
        try:
            text = str(text, "utf-8", errors="replace")  # type: ignore[arg-type]
        except Exception:
            text = repr(text)
    if not text:
        return [text]
    try:
        text = text.replace("\x00", "")
    except Exception:
        pass
    try:
        text = text.encode("utf-8", errors="replace").decode("utf-8", errors="replace")
    except Exception:
        pass

    versions: list[str] = [text]
    cleaned = text
    try:
        cleaned = re.sub(r"[\u200b\u200c\u200d\u2060\ufeff]", "", text)
        if cleaned != text:
            versions.append(cleaned)
    except Exception:
        cleaned = text
    try:
        normalized = unicodedata.normalize("NFKC", cleaned)
        if normalized != cleaned:
            versions.append(normalized)
    except Exception:
        pass
    deduped: list[str] = []
    seen: set[str] = set()
    for item in versions:
        try:
            if item not in seen:
                seen.add(item)
                deduped.append(item)
        except Exception:
            continue
    return deduped or [text]


class PiiDetector:
    """Detect PII and sensitive data in text and structured data."""

    def __init__(
        self,
        patterns: dict[str, tuple[re.Pattern[str], str, str]] | None = None,
        custom_patterns: dict[str, tuple[re.Pattern[str], str, str]] | None = None,
        ignore_patterns: list[str] | None = None,
        severity_threshold: str = "medium",
        use_fast_matching: bool = True,
    ):
        self._patterns = dict(patterns or PII_PATTERNS)
        if custom_patterns:
            self._patterns.update(custom_patterns)
        self._ignored = set(ignore_patterns or [])
        self._threshold = _SEVERITY_RANK.get(severity_threshold, _SEVERITY_RANK["medium"])
        self._use_fast_matching = bool(use_fast_matching)
        self._fast_detector = FastPIIDetector(self._patterns)

    def _scan_text_sequential(self, text: str) -> list[dict[str, Any]]:
        findings: list[dict[str, Any]] = []
        for pattern_name, (compiled, severity, description) in self._patterns.items():
            if pattern_name in self._ignored:
                continue
            if _SEVERITY_RANK.get(severity, 0) < self._threshold:
                continue
            for match in compiled.finditer(text):
                raw = match.group(0)
                findings.append(
                    {
                        "pattern": pattern_name,
                        "severity": severity,
                        "description": description,
                        "match": self._mask(raw, pattern_name),
                        "position": match.start(),
                    }
                )
        return findings

    def scan_text(self, text: str) -> list[dict[str, Any]]:
        text = sanitize_text(text)
        if text is None:
            return []
        all_findings: list[dict[str, Any]] = []
        try:
            versions = preprocess_for_pii(text)
        except Exception:
            return []
        for version in versions:
            try:
                if self._use_fast_matching:
                    all_findings.extend(
                        self._fast_detector.scan(
                            version,
                            threshold_rank=self._threshold,
                            ignored_patterns=self._ignored,
                            mask_fn=self._mask,
                            severity_rank=_SEVERITY_RANK,
                        )
                    )
                else:
                    all_findings.extend(self._scan_text_sequential(version))
            except Exception:
                # Defensive path for malformed Unicode / fuzzed inputs.
                try:
                    all_findings.extend(self._scan_text_sequential(version))
                except Exception:
                    continue
        deduped: list[dict[str, Any]] = []
        seen: set[tuple[str, str, int]] = set()
        for finding in sorted(all_findings, key=lambda item: int(item.get("position", 0))):
            signature = (
                str(finding.get("pattern", "")),
                str(finding.get("match", "")),
                int(finding.get("position", 0)),
            )
            if signature in seen:
                continue
            seen.add(signature)
            deduped.append(finding)
        return deduped

    def scan_dict(self, data: dict[str, Any], path: str = "") -> list[dict[str, Any]]:
        findings: list[dict[str, Any]] = []
        for item_path, value in self._iter_strings(data, path):
            for finding in self.scan_text(value):
                enriched = dict(finding)
                enriched["path"] = item_path
                findings.append(enriched)
        return findings

    def classify_data(self, text: str) -> str:
        findings = self.scan_text(text)
        if any(item["severity"] == "critical" for item in findings):
            return "restricted"
        if any(item["severity"] == "high" for item in findings):
            return "confidential"
        if any(item["severity"] == "medium" for item in findings):
            return "internal"
        return "public"

    def redact_text(self, text: str) -> str:
        matches: list[tuple[int, int, str]] = []
        for pattern_name, (compiled, severity, _) in self._patterns.items():
            if pattern_name in self._ignored:
                continue
            if _SEVERITY_RANK.get(severity, 0) < self._threshold:
                continue
            for match in compiled.finditer(text):
                matches.append((match.start(), match.end(), pattern_name))

        if not matches:
            return text

        matches.sort(key=lambda item: (item[0], -(item[1] - item[0])))
        merged: list[tuple[int, int, str]] = []
        cursor = -1
        for start, end, pattern_name in matches:
            if start < cursor:
                continue
            merged.append((start, end, pattern_name))
            cursor = end

        chunks: list[str] = []
        position = 0
        for start, end, pattern_name in merged:
            chunks.append(text[position:start])
            chunks.append(f"[REDACTED-{pattern_name}]")
            position = end
        chunks.append(text[position:])
        return "".join(chunks)

    def scan_tool_call(self, tool_name: str, params: dict[str, Any]) -> list[dict[str, Any]]:
        findings = self.scan_dict(params, path="params")
        for finding in findings:
            finding["tool"] = tool_name
        return findings

    def _iter_strings(self, value: Any, path: str) -> list[tuple[str, str]]:
        if isinstance(value, str):
            return [(path or "$", value)]
        if isinstance(value, dict):
            out: list[tuple[str, str]] = []
            for key, item in value.items():
                child = f"{path}.{key}" if path else str(key)
                out.extend(self._iter_strings(item, child))
            return out
        if isinstance(value, list):
            out = []
            for index, item in enumerate(value):
                child = f"{path}[{index}]"
                out.extend(self._iter_strings(item, child))
            return out
        return []

    def _mask(self, raw: str, pattern_name: str) -> str:
        if pattern_name in _CRITICAL_PATTERNS:
            suffix = raw[-4:] if len(raw) >= 4 else raw
            return f"***{suffix}"
        if "@" in raw:
            local, _, domain = raw.partition("@")
            local_masked = (local[:2] + "***") if local else "***"
            return f"{local_masked}@{domain}"
        if len(raw) <= 4:
            return "***"
        return f"{raw[:2]}***{raw[-2:]}"


class PIIDetectorHandler:
    """Backward-compatible rule handler for plugin registry integration."""

    def evaluate(self, rule, request, **kwargs):  # noqa: ANN001, ANN003
        _ = kwargs
        detector = PiiDetector(
            ignore_patterns=rule.get("ignore_patterns")
            if isinstance(rule.get("ignore_patterns"), list)
            else None,
            severity_threshold=rule.get("severity_threshold", "medium"),
        )
        params = request.get("params")
        findings = detector.scan_tool_call(
            request.get("tool") if isinstance(request.get("tool"), str) else "__unknown__",
            params if isinstance(params, dict) else {},
        )
        reasons = [f"pii_detector: {item['pattern']} found in {item.get('path', 'params')}" for item in findings]
        return reasons, ["pii_detector"]


class PiiDetectorPlugin:
    """Orchesis policy plugin for blocking and redacting PII."""

    def __init__(self, config: dict[str, Any] | None = None):
        cfg = config or {}
        self.check_params = bool(cfg.get("check_params", True))
        self.check_responses = bool(cfg.get("check_responses", True))
        self.block_on_pii = bool(cfg.get("block_on_pii", True))
        self.redact_in_logs = bool(cfg.get("redact_in_logs", True))
        self.allowed_pii_tools = {
            item for item in cfg.get("allowed_pii_tools", []) if isinstance(item, str)
        }
        self.detector = PiiDetector(
            ignore_patterns=cfg.get("ignore_patterns")
            if isinstance(cfg.get("ignore_patterns"), list)
            else None,
            severity_threshold=cfg.get("severity_threshold", "medium"),
        )

    def evaluate(
        self,
        tool_name: str,
        params: dict[str, Any],
        agent_id: str,
        context: dict[str, Any],
    ) -> dict[str, Any]:
        _ = (agent_id, context)
        findings = self.detector.scan_tool_call(tool_name, params) if self.check_params else []
        classification = self.detector.classify_data(json_dump_safe(params))
        if findings and self.block_on_pii and tool_name not in self.allowed_pii_tools:
            first = findings[0]
            reason = f"PII detected: {first['pattern']} in {first.get('path', 'params')}"
            return {
                "allowed": False,
                "reasons": [reason],
                "pii_findings": findings,
                "classification": classification,
            }
        return {
            "allowed": True,
            "reasons": [],
            "pii_findings": findings,
            "classification": classification,
        }

    def scan_response(self, tool_name: str, response: str) -> list[dict[str, Any]]:
        _ = tool_name
        if not self.check_responses:
            return []
        return self.detector.scan_text(response)

    def redact_for_audit(self, event: dict[str, Any]) -> dict[str, Any]:
        if not self.redact_in_logs:
            return deepcopy(event)
        redacted = deepcopy(event)
        return _redact_structure(redacted, self.detector)


def _redact_structure(value: Any, detector: PiiDetector) -> Any:
    if isinstance(value, str):
        return detector.redact_text(value)
    if isinstance(value, dict):
        return {key: _redact_structure(item, detector) for key, item in value.items()}
    if isinstance(value, list):
        return [_redact_structure(item, detector) for item in value]
    return value


def json_dump_safe(value: Any) -> str:
    try:
        import json

        return json.dumps(value, ensure_ascii=False, sort_keys=True)
    except Exception:
        return str(value)


PLUGIN_INFO = PluginInfo(
    name="pii_detector",
    rule_type="pii_detector",
    version="2.0",
    description="Detect and block tool calls containing sensitive PII",
    handler=PIIDetectorHandler(),
)



from __future__ import annotations

import json
from pathlib import Path

from orchesis.scanner import McpConfigScanner


def _scan_server(tmp_path: Path, server: dict) -> list:
    path = tmp_path / "mcp.json"
    payload = {"mcpServers": {"s1": server}}
    path.write_text(json.dumps(payload), encoding="utf-8")
    return McpConfigScanner().scan(str(path)).findings


def _has(findings: list, category: str, severity: str | None = None, needle: str | None = None) -> bool:
    for item in findings:
        if item.category != category:
            continue
        if severity is not None and item.severity != severity:
            continue
        if needle is not None and needle not in item.evidence and needle not in item.description:
            continue
        return True
    return False


def test_cve_vulnerable_package(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "npx", "args": ["mcp-remote@0.1.15"], "auth": "token"})
    assert _has(findings, "supply_chain_cve", "critical", "CVE-2025-6514")


def test_cve_fixed_version_no_finding(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "npx", "args": ["mcp-remote@0.1.17"], "auth": "token"})
    assert not _has(findings, "supply_chain_cve")


def test_malicious_package_blocked(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "npx", "args": ["mcp-server-free"], "auth": "token"})
    assert _has(findings, "malicious_package", "critical", "mcp-server-free")


def test_typosquatting_detected(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "npx", "args": ["mcp-server-filesytem"], "auth": "token"})
    assert _has(findings, "typosquatting", "high")


def test_typosquatting_exact_match_no_finding(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "npx", "args": ["mcp-server-filesystem"], "auth": "token"})
    assert not _has(findings, "typosquatting")


def test_unsafe_npx_y(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "npx", "args": ["-y", "mcp-server-fetch"], "auth": "token"})
    assert _has(findings, "supply_chain", "high", "npx -y")


def test_npx_y_with_version_no_finding(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "npx", "args": ["-y", "mcp-server-fetch@1.0.0"], "auth": "token"})
    assert not _has(findings, "supply_chain", "high", "npx -y")


def test_entropy_secret_in_args(tmp_path: Path) -> None:
    token = "A9f2K1x0Qw7Lm3N8Zp4Rt6Yv2Hb9Jk3"
    findings = _scan_server(tmp_path, {"command": "node", "args": [token], "auth": "token"})
    assert _has(findings, "secret_leak", "high", "High-entropy")


def test_low_entropy_no_finding(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "node", "args": ["normal-argument-123"], "auth": "token"})
    assert not _has(findings, "secret_leak", "high", "High-entropy")


def test_aws_key_in_env(tmp_path: Path) -> None:
    findings = _scan_server(
        tmp_path,
        {"command": "node", "args": ["app"], "env": {"AWS_ACCESS_KEY_ID": "AKIA1234567890ABCDEF"}, "auth": "token"},
    )
    assert _has(findings, "secret_leak", "critical", "AWS Access Key")


def test_github_pat_in_args(tmp_path: Path) -> None:
    findings = _scan_server(
        tmp_path,
        {"command": "node", "args": ["ghp_abcdefghijklmnopqrstuvwxyz1234567890"], "auth": "token"},
    )
    assert _has(findings, "secret_leak", "critical", "GitHub PAT")


def test_anthropic_key_detected(tmp_path: Path) -> None:
    findings = _scan_server(
        tmp_path,
        {"command": "node", "args": ["sk-ant-abcdefghijklmnopqrstuvwxyz-abcdefghijklmnopqrstuvwxyz-1234"], "auth": "token"},
    )
    assert _has(findings, "secret_leak", "critical", "Anthropic API Key")


def test_jwt_detected(tmp_path: Path) -> None:
    jwt = "ey" + "A" * 25 + "." + "B" * 25 + "." + "C" * 25
    findings = _scan_server(tmp_path, {"command": "node", "args": [jwt], "auth": "token"})
    assert _has(findings, "secret_leak", "high", "JWT Token")


def test_privileged_mode(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "--privileged"], "auth": "token"})
    assert _has(findings, "docker_security", "critical", "--privileged")


def test_docker_socket_mount(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "-v", "/var/run/docker.sock:/var/run/docker.sock"], "auth": "token"})
    assert _has(findings, "docker_security", "critical", "/var/run/docker.sock")


def test_sensitive_ssh_mount(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "-v", "~/.ssh:/root/.ssh:ro"], "auth": "token"})
    assert _has(findings, "docker_security", "critical", "~/.ssh")


def test_network_host(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "--network=host"], "auth": "token"})
    assert _has(findings, "docker_security", "high", "--network=host")


def test_mount_without_ro(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "-v", "~/.aws:/root/.aws"], "auth": "token"})
    assert _has(findings, "docker_security", "high", "read-only")


def test_running_as_root(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "alpine:3.20"], "auth": "token"})
    assert _has(findings, "docker_security", "medium", "runs as root")


def test_no_resource_limits(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "--user", "1000:1000", "alpine:3.20"], "auth": "token"})
    assert _has(findings, "docker_security", "medium", "resource limits")


def test_seccomp_unconfined(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "--security-opt", "seccomp=unconfined"], "auth": "token"})
    assert _has(findings, "docker_security", "high", "unconfined")


def test_dangerous_cap_add(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "--cap-add", "SYS_ADMIN"], "auth": "token"})
    assert _has(findings, "docker_security", "high", "SYS_ADMIN")


def test_image_no_digest(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"command": "docker", "args": ["run", "--user", "1000:1000", "--memory=256m"], "image": "docker.io/library/alpine", "auth": "token"})
    assert _has(findings, "docker_security", "medium", "digest or tag")


def test_clean_config_no_docker_findings(tmp_path: Path) -> None:
    findings = _scan_server(
        tmp_path,
        {
            "command": "docker",
            "args": [
                "run",
                "--user",
                "1000:1000",
                "--memory=256m",
                "--cpus=1",
                "--pids-limit=128",
                "-v",
                "/workspace:/workspace:ro",
                "--cap-add",
                "CHOWN",
                "alpine:3.20",
            ],
            "auth": "token",
            "docker": {"user": "1000:1000"},
            "image": "docker.io/library/alpine:3.20@sha256:abc",
        },
    )
    assert not any(item.category == "docker_security" and item.severity in {"critical", "high", "medium"} for item in findings)


def test_existing_checks_still_pass(tmp_path: Path) -> None:
    findings = _scan_server(tmp_path, {"url": "http://0.0.0.0:8080", "tools": ["*"]})
    categories = {item.category for item in findings}
    assert "binding_exposure" in categories
    assert "no_auth" in categories
    assert "dangerous_tools" in categories




"""Simple deterministic agent loop integrated with Orchesis decisions."""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from orchesis.agent.planner import load_task_catalog, resolve_task_steps
from orchesis.config import load_agent_registry, load_policy
from orchesis.engine import evaluate
from orchesis.logger import append_decision
from orchesis.models import Decision
from orchesis.state import RateLimitTracker


@dataclass
class AgentState:
    task: str
    steps: list[dict[str, Any]] = field(default_factory=list)
    current_step: int = 0
    max_steps: int = 10
    status: str = "running"  # running | completed | blocked | failed


class SimpleAgent:
    def __init__(
        self,
        policy_path: str,
        tools: list[dict[str, Any]],
        *,
        tasks_path: str | Path = "examples/agent_tasks.yaml",
        log_path: str | Path = "decisions.jsonl",
    ):
        self.policy = load_policy(policy_path)
        has_identity_config = "agents" in self.policy or "default_trust_tier" in self.policy
        self.registry = load_agent_registry(self.policy) if has_identity_config else None
        self.tools = tools
        self.tasks_path = Path(tasks_path)
        self.log_path = Path(log_path)
        self.catalog = load_task_catalog(self.tasks_path)

    def run(self, task: str, *, max_steps: int = 10) -> AgentState:
        """Main loop: plan -> verify -> execute -> repeat."""
        state = AgentState(task=task, max_steps=max_steps)
        tracker = RateLimitTracker(persist_path=None)

        while state.current_step < state.max_steps:
            action = self.plan_next_action(state)
            if action is None:
                state.status = "completed"
                break

            decision = self.verify_action(action, tracker=tracker)
            effective_agent = (
                action.get("context", {}).get("agent")
                if isinstance(action.get("context"), dict)
                and isinstance(action.get("context", {}).get("agent"), str)
                else "cursor"
            )
            identity = self.registry.get(effective_agent) if self.registry is not None else None
            step_record: dict[str, Any] = {
                "step": state.current_step + 1,
                "tool": action["tool"],
                "params": action.get("params", {}),
                "agent": effective_agent,
                "trust_tier": identity.trust_tier.name.lower() if identity is not None else "n/a",
                "cost": action.get("cost", 0.0),
                "decision": "ALLOW" if decision.allowed else "DENY",
                "reasons": decision.reasons,
            }

            request = {
                "tool": action["tool"],
                "params": action.get("params", {}),
                "cost": action.get("cost", 0.0),
                "context": {
                    "task": state.task,
                    "agent": "cursor",
                    **(action.get("context") if isinstance(action.get("context"), dict) else {}),
                },
            }
            append_decision(decision, request, self.log_path)

            if decision.allowed:
                execution = self.execute_action(action)
                step_record["result"] = execution
            else:
                self.handle_block(state, decision)

            state.steps.append(step_record)
            state.current_step += 1

        if state.status == "running":
            # Max steps reached before completing planned sequence.
            state.status = "failed"

        return state

    def plan_next_action(self, state: AgentState) -> dict[str, Any] | None:
        """Deterministic mapping: task + history -> next action."""
        task_steps = resolve_task_steps(state.task, self.catalog)
        if state.current_step >= len(task_steps):
            return None
        action = task_steps[state.current_step]
        if not isinstance(action, dict) or "tool" not in action:
            raise ValueError(f"Invalid task step at index {state.current_step}")
        return action

    def verify_action(self, action: dict[str, Any], *, tracker: RateLimitTracker) -> Decision:
        """Pass action through Orchesis evaluate()."""
        request = {
            "tool": action["tool"],
            "params": action.get("params", {}),
            "cost": action.get("cost", 0.0),
            "context": {
                "agent": "cursor",
                **(action.get("context") if isinstance(action.get("context"), dict) else {}),
            },
        }
        return evaluate(request, self.policy, state=tracker, registry=self.registry)

    def execute_action(self, action: dict[str, Any]) -> dict[str, Any]:
        """Execute allowed tool call via deterministic stubs."""
        tool = action["tool"]
        params = action.get("params", {})

        if tool == "read_file":
            path = params.get("path", "")
            return {"status": "ok", "content": f"stub content from {path}"}
        if tool == "delete_file":
            return {"status": "ok", "message": "deleted"}
        if tool == "run_sql":
            query = str(params.get("query", ""))
            if query.strip().upper().startswith("SELECT SUM"):
                return {"status": "ok", "result": "1000"}
            return {"status": "ok", "result": "executed"}
        if tool == "write_file":
            return {"status": "ok", "message": "written"}
        if tool == "api_call":
            endpoint = params.get("endpoint", "unknown")
            return {"status": "ok", "message": f"called {endpoint}"}

        return {"status": "ok", "message": f"tool {tool} executed"}

    def handle_block(self, state: AgentState, decision: Decision) -> None:
        """Handle blocked action. Current strategy: record and continue."""
        _ = decision
        state.status = "running"



"""Shared primitives for alert integrations."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any


_SEVERITY_ORDER = {"low": 0, "medium": 1, "high": 2, "critical": 3}


@dataclass
class AlertEvent:
    """Normalized alert event delivered to external integrations."""

    action: str
    severity: str
    agent_id: str
    rule_id: str
    pattern: str
    description: str
    remediation: str
    timestamp: str = field(default_factory=lambda: datetime.now(timezone.utc).isoformat())
    metadata: dict[str, Any] = field(default_factory=dict)


class BaseIntegration:
    """Base class for all alert integrations."""

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        self.config = config if isinstance(config, dict) else {}

    def should_alert(self, event: AlertEvent, config: dict[str, Any] | None = None) -> bool:
        """Check whether event passes integration filters."""
        cfg = config if isinstance(config, dict) else self.config
        if not bool(cfg.get("enabled", False)):
            return False
        on_events = cfg.get("on", [])
        if isinstance(on_events, list) and on_events:
            allowed = {str(item).strip().lower() for item in on_events if str(item).strip()}
            if str(event.action).strip().lower() not in allowed:
                return False
        min_severity = str(cfg.get("min_severity", "low")).strip().lower()
        event_rank = _SEVERITY_ORDER.get(str(event.severity).strip().lower(), 0)
        min_rank = _SEVERITY_ORDER.get(min_severity, 0)
        return event_rank >= min_rank

    def send(self, event: AlertEvent) -> Any:  # pragma: no cover - interface only
        raise NotImplementedError




from __future__ import annotations

import json
from pathlib import Path

from orchesis.scanner import McpConfigScanner


def _scan_servers(tmp_path: Path, servers: dict) -> list:
    path = tmp_path / "mcp.json"
    path.write_text(json.dumps({"mcpServers": servers}), encoding="utf-8")
    return McpConfigScanner().scan(str(path)).findings


def _scan_one(tmp_path: Path, server: dict) -> list:
    return _scan_servers(tmp_path, {"s1": server})


def _has(findings: list, category: str, severity: str | None = None, needle: str | None = None) -> bool:
    for item in findings:
        if item.category != category:
            continue
        if severity is not None and item.severity != severity:
            continue
        if needle is not None and needle not in item.description and needle not in item.evidence:
            continue
        return True
    return False


def test_sse_http_no_tls(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"url": "http://example.com/sse", "transport": "sse", "auth": "token"})
    assert _has(findings, "transport_security", "high")


def test_sse_https_no_finding(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"url": "https://example.com/sse", "transport": "sse", "auth": "token"})
    assert not _has(findings, "transport_security")


def test_cross_server_exfil(tmp_path: Path) -> None:
    findings = _scan_servers(
        tmp_path,
        {
            "fs": {"command": "mcp-server-filesystem", "auth": "token"},
            "net": {"command": "mcp-server-fetch", "auth": "token"},
        },
    )
    assert _has(findings, "exfiltration_risk", "medium")


def test_single_server_no_exfil(tmp_path: Path) -> None:
    findings = _scan_servers(tmp_path, {"fs": {"command": "mcp-server-filesystem", "auth": "token"}})
    assert not _has(findings, "exfiltration_risk")


def test_shell_server_name(tmp_path: Path) -> None:
    findings = _scan_servers(tmp_path, {"shell-access": {"command": "node server.js", "auth": "token"}})
    assert _has(findings, "shell_execution", "high")


def test_terminal_command(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"command": "terminal-bridge", "auth": "token"})
    assert _has(findings, "shell_execution", "high")


def test_exec_pattern(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"command": "exec-server", "auth": "token"})
    assert _has(findings, "shell_execution", "high")


def test_sudo_in_args(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"command": "bash", "args": ["sudo", "cat", "/etc/hosts"], "auth": "token"})
    assert _has(findings, "privilege_escalation", "critical")


def test_doas_in_args(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"command": "bash", "args": ["doas", "cat", "/etc/hosts"], "auth": "token"})
    assert _has(findings, "privilege_escalation", "critical")


def test_path_traversal_args(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"command": "node", "args": ["../../../etc/passwd"], "auth": "token"})
    assert _has(findings, "path_traversal", "high")


def test_windows_path_traversal(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"command": "node", "args": ["..\\..\\windows\\system32"], "auth": "token"})
    assert _has(findings, "path_traversal", "high")


def test_wildcard_path(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"command": "node", "args": ["/data/*"], "auth": "token"})
    assert _has(findings, "file_access", "medium", "Wildcard glob")


def test_glob_double_star(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"command": "node", "args": ["/home/**"], "auth": "token"})
    assert _has(findings, "file_access", "medium", "Wildcard glob")


def test_cors_missing_remote(tmp_path: Path) -> None:
    findings = _scan_one(tmp_path, {"url": "https://api.example.com/mcp", "auth": "token"})
    assert _has(findings, "cors_missing", "medium")


def test_cors_present_no_finding(tmp_path: Path) -> None:
    findings = _scan_one(
        tmp_path,
        {"url": "https://api.example.com/mcp", "auth": "token", "cors": {"allow": ["https://app.example.com"]}},
    )
    assert not _has(findings, "cors_missing")


def test_excessive_servers(tmp_path: Path) -> None:
    servers = {f"s{i}": {"command": "node", "auth": "token"} for i in range(11)}
    findings = _scan_servers(tmp_path, servers)
    assert _has(findings, "attack_surface", "medium", "11")


def test_ten_servers_no_finding(tmp_path: Path) -> None:
    servers = {f"s{i}": {"command": "node", "auth": "token"} for i in range(10)}
    findings = _scan_servers(tmp_path, servers)
    assert not _has(findings, "attack_surface")


def test_shared_credentials(tmp_path: Path) -> None:
    shared = "TokenValue1234567890"
    findings = _scan_servers(
        tmp_path,
        {
            "a": {"command": "node", "auth": "token", "env": {"API_TOKEN": shared}},
            "b": {"command": "node", "auth": "token", "env": {"SERVICE_TOKEN": shared}},
        },
    )
    assert _has(findings, "credential_sharing", "high")


def test_unique_credentials_no_finding(tmp_path: Path) -> None:
    findings = _scan_servers(
        tmp_path,
        {
            "a": {"command": "node", "auth": "token", "env": {"API_TOKEN": "TokenValue1234567890"}},
            "b": {"command": "node", "auth": "token", "env": {"SERVICE_TOKEN": "OtherValue1234567899"}},
        },
    )
    assert not _has(findings, "credential_sharing")


def test_phase1_phase2_unaffected(tmp_path: Path) -> None:
    findings = _scan_one(
        tmp_path,
        {"command": "npx", "args": ["mcp-remote@0.1.15"], "auth": "token", "autoApprove": True},
    )
    assert _has(findings, "supply_chain_cve", "critical")
    assert _has(findings, "permissions", "high")




"""Prometheus-compatible metrics collector."""

from __future__ import annotations

import threading
from collections import defaultdict
from typing import Any

from orchesis.telemetry import DecisionEvent, EventEmitter


class MetricsCollector(EventEmitter):
    """Collects metrics from decision events."""

    def __init__(self):
        self._counters: dict[str, int] = defaultdict(int)
        self._histograms: dict[str, list[float]] = defaultdict(list)
        self._lock = threading.Lock()

    def emit(self, event: DecisionEvent) -> None:
        with self._lock:
            self._counters[f"orchesis_decisions_total|decision={event.decision}"] += 1
            self._counters[
                f"orchesis_decisions_by_agent|agent={event.agent_id}|decision={event.decision}"
            ] += 1
            self._counters[
                f"orchesis_decisions_by_tool|tool={event.tool}|decision={event.decision}"
            ] += 1
            if any(reason.startswith("anomaly:") for reason in event.reasons):
                self._counters["orchesis_anomalies_total"] += 1
            self._histograms["orchesis_evaluation_duration_us"].append(
                float(event.evaluation_duration_us)
            )

    def _counter_value(self, key: str) -> int:
        return int(self._counters.get(key, 0))

    def prometheus_text(self) -> str:
        with self._lock:
            lines: list[str] = []
            lines.append("# HELP orchesis_decisions_total Total decisions")
            lines.append("# TYPE orchesis_decisions_total counter")
            for decision in ("ALLOW", "DENY"):
                value = self._counter_value(f"orchesis_decisions_total|decision={decision}")
                lines.append(f'orchesis_decisions_total{{decision="{decision}"}} {value}')

            lines.append("# HELP orchesis_decisions_by_agent Decisions by agent and decision")
            lines.append("# TYPE orchesis_decisions_by_agent counter")
            for key, value in sorted(self._counters.items()):
                if not key.startswith("orchesis_decisions_by_agent|"):
                    continue
                parts = key.split("|")
                agent = parts[1].split("=", 1)[1]
                decision = parts[2].split("=", 1)[1]
                lines.append(
                    f'orchesis_decisions_by_agent{{agent="{agent}",decision="{decision}"}} {value}'
                )

            lines.append("# HELP orchesis_decisions_by_tool Decisions by tool and decision")
            lines.append("# TYPE orchesis_decisions_by_tool counter")
            for key, value in sorted(self._counters.items()):
                if not key.startswith("orchesis_decisions_by_tool|"):
                    continue
                parts = key.split("|")
                tool = parts[1].split("=", 1)[1]
                decision = parts[2].split("=", 1)[1]
                lines.append(
                    f'orchesis_decisions_by_tool{{tool="{tool}",decision="{decision}"}} {value}'
                )

            lines.append("# HELP orchesis_anomalies_total Total detected anomalies")
            lines.append("# TYPE orchesis_anomalies_total counter")
            lines.append(
                f"orchesis_anomalies_total {self._counter_value('orchesis_anomalies_total')}"
            )

            values = self._histograms.get("orchesis_evaluation_duration_us", [])
            total = sum(values)
            count = len(values)
            lines.append("# HELP orchesis_evaluation_duration_us Evaluation latency")
            lines.append("# TYPE orchesis_evaluation_duration_us histogram")
            lines.append(f"orchesis_evaluation_duration_us_sum {total}")
            lines.append(f"orchesis_evaluation_duration_us_count {count}")
            return "\n".join(lines) + "\n"

    def reset(self) -> None:
        with self._lock:
            self._counters = defaultdict(int)
            self._histograms = defaultdict(list)

    def snapshot(self) -> dict[str, Any]:
        with self._lock:
            return {
                "counters": dict(self._counters),
                "histograms": {key: list(values) for key, values in self._histograms.items()},
            }
